"use strict";

exports.__esModule = true;
exports.cleanEmptyExprs = exports.forEach = exports.map = exports.flatMap = exports.filter = exports.pack = exports.packed = exports.enableIf = exports.select = exports.adjustFieldTypeSimple = exports.adjustFieldType = exports.makeStmt = exports.makeExpr = exports.transform = exports.babelBridge = exports.checkpoint = exports.checkpointLazy = exports.wrap = exports.UpdateOpts = exports.Opts = exports.find = exports.Subst = exports.OutputStream = exports.NoInput = exports.AutoArrStream = exports.AutoStream = exports.LevelArrStream = exports.LevelStream = exports.LookaheadStream = exports.LookaheadArrStream = exports.Output = exports.ArrayLookahead = exports.Lookahead = exports.ExtIterator = undefined;
exports.optsScope = optsScope;
exports.optsScopeLift = optsScopeLift;
exports.setOpts = setOpts;
exports.Level = Level;
exports.WithPeel = WithPeel;
exports.parse = parse;
exports.toks = toks;
exports.Template = Template;
exports.Stream = Stream;
exports.lookahead = lookahead;
exports.levels = levels;
exports.auto = auto;
exports.output = output;
exports.skip = skip;
exports.setType = setType;
exports.setPos = setPos;
exports.repos = repos;
exports.reposOne = reposOne;
exports.reposOneArr = reposOneArr;
exports.completeSubst = completeSubst;
exports.toArray = toArray;
exports.result = result;
exports.tillLevel = tillLevel;
exports.toBlockBody = toBlockBody;
exports.inBlockBody = inBlockBody;
exports.fileBody = fileBody;
exports.hasAnnot = hasAnnot;
exports.clone = clone;
exports.till = till;
exports.concat = concat;
exports.share = share;
exports.stage = stage;
exports.babelPreset = babelPreset;
exports.tee = tee;
exports.makeExprPass = makeExprPass;
exports.removeEmptyBlocks = removeEmptyBlocks;
exports.cons = cons;
exports.la = la;
exports.unpack = unpack;
exports.reverse = reverse;
exports.mapPush = mapPush;
exports.mapAdd = mapAdd;
exports.groupWith /*::<K,V,W>*/ = groupWith;
exports.group /*::<K,V>*/ = group;
exports.groupUniq /*::<K,V>*/ = groupUniq;
exports.pipe = pipe;
exports.curryN = curryN;
exports.curry = curry;

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

var _core = require("./core");

var _babelTypes = require("babel-types");

var T = _interopRequireWildcard(_babelTypes);

var _babylon = require("babylon");

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

var M = require("@effectful/es-loose-rt");

const BROWSER_DEBUG = typeof window !== "undefined" && window.chrome;
let _opts = {};

/** runs `fun` and reset global options after its exit */
function optsScope(fun) {
  const save = _opts;
  try {
    return fun();
  } finally {
    _opts = save;
  }
}

/** returns function calling `fun` and reseting global options after exit */
function optsScopeLift(fun) {
  return function () {
    const save = _opts;
    try {
      return fun.apply(this, arguments);
    } finally {
      _opts = save;
    }
  };
}

/** sets global options */
function setOpts(opts) {
  _opts = opts;
}

/**
 * adds `take` function to ES6 iterators interface
 * children classes may implement one of `take` or `next` methods
 */
class ExtIterator {
  [Symbol.iterator]() {
    return this;
  }
  /**
   * ES6 Iterator interface `next`
   */
  next(v) {
    const c = this.take(v);
    return { value: c, done: c == null };
  }
  /**
   * same as `next` but returns either next value or null if done
   */
  take(v) {
    const c = this.next(v);
    return c.done ? null : c.value;
  }
  constructor(cont) {
    this._cont = cont;
  }
}

exports.ExtIterator = ExtIterator; /** 
                                    * iterators wrapper keeping a single element lookahead, which may be accessed
                                    * with `cur` method
                                    *
                                    * the iterator may be shared
                                    */

class Lookahead extends ExtIterator {
  constructor(cont) {
    super(cont);
    this._inner = cont[Symbol.iterator]();
    let i = this._inner.next();
    assert.ok(!i.done, "input iterator should be not-empty");
    this.first = i.value;
    this.opts = this.first.value && this.first.value.opts || _opts;
    this._cur = i;
  }
  next(v) {
    const cur = this._cur;
    if (!cur.done) {
      if (cur.value.value != null && cur.value.value.opts != null) this.opts = cur.value.value.opts;
      this._cur = this._inner.next(v);
    }
    return cur;
  }
  take(v) {
    const cur = this._cur;
    if (cur.done) return null;
    if (cur.value.opts != null) this.opts = cur.value.opts;
    this._cur = this._inner.next(v);
    return cur.value;
  }
  cur() {
    return this._cur.done ? undefined : this._cur.value;
  }
}

exports.Lookahead = Lookahead; /** same as `Lookahead` but optimized for Array's input */

class ArrayLookahead extends ExtIterator {
  constructor(cont) {
    super(cont);
    assert.ok(cont.length > 0, "input iterator should be not-empty");
    this._x = 0;
    this.first = cont[0];
    this.opts = this.first.value && this.first.value.opts || _opts;
  }
  next(v) {
    const c = this._cont[this._x++];
    if (c != null && c.value != null && c.value.opts != null) this.opts = c.value.opts;
    return { value: c, done: c == null };
  }
  take(v) {
    const c = this._cont[this._x++];
    if (c != null && c.opts != null) this.opts = c.opts;
    return c;
  }
  cur() {
    return this._cont[this._x];
  }
}

exports.ArrayLookahead = ArrayLookahead;
const ctrlTok = Symbol("ctrlTok");
const ctrlTokGen = Symbol("ctrlTokGen");
const storedTok = Symbol("storedTok");

/** iterator's wrapper proiding tokens output method */
const Output = exports.Output = Super => class Output extends Super {
  constructor(cont) {
    super(cont);
    this._stack = [];
  }
  /** smart constructor for a token */
  valCtor(pos, type, value) {
    let node = null;
    if (value == null) {
      if (type != null && typeof type !== "symbol") {
        if (type.node != null) {
          value = type;
          node = value.node;
          type = null;
        } else if (type.type != null) {
          node = type;
          value = { node };
          type = null;
        } else if (value == null) {
          value = type;
          type = null;
        }
      }
      if (value == null) value = {};
    } else node = value.node;
    if (type == null) {
      if (value != null && value.typeInfo != null) {
        type = value.typeInfo.sym;
      } else if (node != null && node.type != null) {
        type = _core.Tag[node.type];
      } else {
        if ((0, _core.symInfo)(pos).kind === "ctrl") type = pos;
      }
    }
    assert.ok(type, "couldn't guess type");
    if (node == null) {
      value.node = node = type === _core.Tag.Array ? [] : type === _core.Tag.Null ? null : {};
    }
    if (!value.opts) value.opts = this.opts;
    return [pos, type, value];
  }
  /** 
   * outputs tokens from parsed template
   * see module `toks` function for details 
   */
  toks(pos, node, ...syms) {
    var ctx = M.generator();
    ctx.$exit = _3;
    ctx.handle = _4;
    ctx.step = _1;
    return ctx.unwrap;

    function _1() {
      var a;

      try {
        a = toks(pos, node, ...syms);
        ctx.step = _2;
        return ctx.delegate(a);
      } catch (ex) {
        ctx.step = _4;
        return ctx.step(ex);
      }
    }

    function _2() {
      try {
        return ctx.pure();
      } catch (ex) {
        return ctx.pure(ex);
      }
    }

    function _3(r) {
      ctx.$exit = _3;
      ctx.handle = _4;

      try {
        return ctx.pure(r);
      } catch (ex) {
        return ctx.pure(ex);
      }
    }

    function _4(e) {
      return ctx.raise(e);
    }
  }
  /** outputs opening tag */
  enter(pos, type, value) {
    [pos, type, value] = this.valCtor(pos, type, value);
    this._stack.unshift({ $: storedTok,
      tok: { enter: false, leave: true,
        pos: pos, type: type, value: value } });
    return { enter: true, leave: false, type, pos, value };
  }
  /** outputs terminal tag */
  tok(pos, type, value) {
    [pos, type, value] = this.valCtor(pos, type, value);
    return { enter: true, leave: true, type, pos, value };
  }
  /** outputs closing tag doing other actions stored in the stack */
  leave() {
    var f,
        ctx = M.generator(),
        _this = this,
        r;

    ctx.$exit = _5;
    ctx.handle = _6;
    ctx.step = _1;
    return ctx.unwrap;

    function _1() {
      exit: {
        try {
          f = void 0;
          ctx.step = _2;
          break exit;
        } catch (ex) {
          ctx.step = _6;
          return ctx.step(ex);
        }
      }

      return ctx.step();
    }

    function _2() {
      var a, b;

      exit: {
        try {
          if (f = _this._stack.shift()) {
            switch (f.$) {
              case ctrlTok:
                if (f.run(_this)) {
                  return ctx.pure(r);
                } else {
                  break exit;
                }

              case ctrlTokGen:
                a = f.run(_this);
                ctx.step = _3;
                return ctx.delegate(a);

              default:
                b = f.tok;
                ctx.step = _4;
                ctx.value = b;
                return ctx;

            }
          } else {
            return ctx.pure(r);
          }
        } catch (ex) {
          ctx.step = _6;
          return ctx.step(ex);
        }
      }

      return ctx.step();
    }

    function _3(a) {
      exit: {
        try {
          if (a) {
            return ctx.pure(r);
          } else {
            ctx.step = _2;
            break exit;
          }
        } catch (ex) {
          ctx.step = _6;
          return ctx.step(ex);
        }
      }

      return ctx.step();
    }

    function _4() {
      try {
        return ctx.pure(f.tok);
      } catch (ex) {
        ctx.step = _6;
        return ctx.step(ex);
      }
    }

    function _5(r) {
      ctx.$exit = _5;
      ctx.handle = _6;

      try {
        return ctx.pure(r);
      } catch (ex) {
        return ctx.pure(ex);
      }
    }

    function _6(e) {
      return ctx.raise(e);
    }
  }
  /** 
   * gets a mark (function) to close all tags opened 
   * after the function is called 
   */
  label() {
    const pos = this._stack.length;
    const t = this;
    return function () {
      var sub,
          i,
          ctx = M.generator(),
          loop;
      ctx.$exit = _3;
      ctx.handle = _4;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        exit: {
          try {
            sub = t._stack.splice(0, t._stack.length - pos);
            loop = M.iterator(sub);
            ctx.step = _2;
            break exit;
          } catch (ex) {
            ctx.step = _4;
            return ctx.step(ex);
          }
        }

        return ctx.step();
      }

      function _2() {
        var a, b;

        exit: {
          try {
            if (!(loop = loop.step()).done) {
              i = loop.value;

              switch (i.$) {
                case ctrlTok:
                  i.run(t);
                  break exit;

                case ctrlTokGen:
                  a = i.run(t);
                  return ctx.delegate(a);

                default:
                  b = i.tok;
                  ctx.value = b;
                  return ctx;

              }
            } else {
              return ctx.pure();
            }
          } catch (ex) {
            ctx.step = _4;
            return ctx.step(ex);
          }
        }

        return ctx.step();
      }

      function _3(r) {
        ctx.$exit = _3;
        ctx.handle = _4;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _4(e) {
        return ctx.raise(e);
      }
    };
  }
};

/** iterator's wrapper adding tree's level tracking functionality */
function Level(Super) {
  function one(t) {
    var c,
        exit,
        i,
        one = M.generator(),
        loop,
        r;
    one.$exit = _4;
    one.handle = _5;
    one.step = _1;
    return one.unwrap;

    function _1() {
      exit: {
        try {
          c = t.cur();

          if (c == null || !c.enter) {
            return one.pure(null);
          } else {
            exit = t.level;

            c.value.stageName = t.first.value.stageName;
            i = void 0;
            loop = M.iterator(t);
            one.step = _2;
            break exit;
          }
        } catch (ex) {
          one.step = _5;
          return one.step(ex);
        }
      }

      return one.step();
    }

    function _2() {
      try {
        if (!(loop = loop.step()).done) {
          i = loop.value;
          one.step = _3;
          one.value = i;
          return one;
        } else {
          return one.pure(i);
        }
      } catch (ex) {
        one.step = _5;
        return one.step(ex);
      }
    }

    function _3() {
      exit: {
        try {
          if (exit >= t.level) {
            return one.pure(i);
          } else {
            one.step = _2;
            break exit;
          }
        } catch (ex) {
          one.step = _5;
          return one.step(ex);
        }
      }

      return one.step();
    }

    function _4(r) {
      one.$exit = _4;
      one.handle = _5;

      try {
        return one.pure(r);
      } catch (ex) {
        return one.pure(ex);
      }
    }

    function _5(e) {
      return one.raise(e);
    }
  }
  function sub(t) {
    var c,
        exit,
        i,
        _c,
        sub = M.generator(),
        loop,
        r;

    sub.$exit = _4;
    sub.handle = _5;
    sub.step = _1;
    return sub.unwrap;

    function _1() {
      exit: {
        try {
          c = t.cur();

          if (c == null || !c.enter) {
            return sub.pure(null);
          } else {
            c.value.stageName = t.first.value.stageName;
            exit = t.level;
            i = void 0;
            loop = M.iterator(t);
            sub.step = _2;
            break exit;
          }
        } catch (ex) {
          sub.step = _5;
          return sub.step(ex);
        }
      }

      return sub.step();
    }

    function _2() {
      try {
        if (!(loop = loop.step()).done) {
          i = loop.value;
          sub.step = _3;
          sub.value = i;
          return sub;
        } else {
          return sub.pure(i);
        }
      } catch (ex) {
        sub.step = _5;
        return sub.step(ex);
      }
    }

    function _3() {
      var fn;

      exit: {
        try {
          if (exit >= t.level) {
            _c = t.cur();

            if (_c == null || !_c.enter || exit > t.level) {
              return sub.pure(i);
            } else {
              sub.step = _2;
              fn = sub.step;
              break exit;
            }
          } else {
            sub.step = _2;
            fn = sub.step;
            break exit;
          }
        } catch (ex) {
          sub.step = _5;
          return sub.step(ex);
        }
      }

      return fn();
    }

    function _4(r) {
      sub.$exit = _4;
      sub.handle = _5;

      try {
        return sub.pure(r);
      } catch (ex) {
        return sub.pure(ex);
      }
    }

    function _5(e) {
      return sub.raise(e);
    }
  }
  return class Level extends Super {
    constructor(cont) {
      super(cont);
      this.level = 0;
    }
    next(v) {
      const c = super.next(v);
      if (c.done) return c;
      const t = c.value;
      if (t.enter) this.level++;
      if (t.leave) this.level--;
      return c;
    }
    take(v) {
      const c = super.take(v);
      if (c == null) return c;
      if (c.enter) this.level++;
      if (c.leave) this.level--;
      return c;
    }
    /** 
     * returns an iterator, if next tag is opening the iterator will 
     * output all tokens until it is closed, or empty otherwise
     */
    one() {
      return one(this);
    }
    /** 
     * returns an iterator to traverse all tokens until exiting from 
     * the current level
     */
    sub() {
      return sub(this);
    }
    /** gets next tag if it is in the current level or null otherwise */
    curLev() {
      const v = this.cur();
      if (!v || !v.enter) return null;
      return v;
    }
    /** 
     * outputs all tokens until token with position `pos` 
     * on the current level, returns the tag if found, it doesn't 
     * remove the tag from the original iterator
     */
    untilPos(pos) {
      var i,
          ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _2;
      ctx.handle = _3;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          if ((i = _this.curLev()) != null && i.pos !== pos) {
            a = one(_this);
            return ctx.delegate(a);
          } else {
            return ctx.pure(i);
          }
        } catch (ex) {
          ctx.step = _3;
          return ctx.step(ex);
        }
      }

      function _2(r) {
        ctx.$exit = _2;
        ctx.handle = _3;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _3(e) {
        return ctx.raise(e);
      }
    }

    /** same as `untilPos` but takes the tag from the original stream */
    findPos(pos) {
      var i,
          ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _3;
      ctx.handle = _4;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          a = _this.untilPos(pos);
          ctx.step = _2;
          return ctx.delegate(a);
        } catch (ex) {
          ctx.step = _4;
          return ctx.step(ex);
        }
      }

      function _2(a) {
        try {
          i = a;

          if (i != null) {
            _this.take();
          }return ctx.pure(i);
        } catch (ex) {
          ctx.step = _4;
          return ctx.step(ex);
        }
      }

      function _3(r) {
        ctx.$exit = _3;
        ctx.handle = _4;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _4(e) {
        return ctx.raise(e);
      }
    }

    /** same as `findPos` but also outputs the found tag */
    toPos(pos) {
      var p,
          ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _4;
      ctx.handle = _5;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          a = _this.findPos(pos);
          ctx.step = _2;
          return ctx.delegate(a);
        } catch (ex) {
          ctx.step = _5;
          return ctx.step(ex);
        }
      }

      function _2(a) {
        try {
          p = a;

          assert.ok(p);
          ctx.step = _3;
          ctx.value = p;
          return ctx;
        } catch (ex) {
          ctx.step = _5;
          return ctx.step(ex);
        }
      }

      function _3() {
        try {
          return ctx.pure(p);
        } catch (ex) {
          ctx.step = _5;
          return ctx.step(ex);
        }
      }

      function _4(r) {
        ctx.$exit = _4;
        ctx.handle = _5;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _5(e) {
        return ctx.raise(e);
      }
    }
  };
}

/** extends iterator with `peel` function (see its doc) */
function WithPeel(Super) {
  // means we are to skip next tag from input because it is in the stack already
  const skipTag = { $: ctrlTok, run(t) {
      t.take();
    }, t: "skip"
    // virtual close (already closed in token)
  };const vCloseTag = { $: ctrlTok, run() {}, t: "close" };
  return class WithPeel extends Super {
    constructor(cont) {
      super(cont);
    }
    /** 
     * - if `i` is undefined reads a token from the stream
     * - makes an opening tag (copied from the read one or i) and returns it
     * - save closing tag into a stack to output it after leaving its level
     */
    peel(i) {
      if (i == null) i = this.take();
      assert.ok(i.enter);
      const res = this.enter(i.pos, i.type, i.value);
      this._stack.unshift(i.leave ? vCloseTag : skipTag);
      return res;
    }
    /** 
     * looks a `pos` tag in the input and returns it, 
     * also output its peel
     */
    peelTo(pos) {
      var i,
          ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _4;
      ctx.handle = _5;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          assert.notEqual(_this._stack[0], vCloseTag);
          a = _this.findPos(pos);
          ctx.step = _2;
          return ctx.delegate(a);
        } catch (ex) {
          ctx.step = _5;
          return ctx.step(ex);
        }
      }

      function _2(a) {
        var b;

        try {
          i = a;

          assert.ok(i);
          b = _this.peel(i);
          ctx.step = _3;
          ctx.value = b;
          return ctx;
        } catch (ex) {
          ctx.step = _5;
          return ctx.step(ex);
        }
      }

      function _3() {
        try {
          return ctx.pure(i);
        } catch (ex) {
          ctx.step = _5;
          return ctx.step(ex);
        }
      }

      function _4(r) {
        ctx.$exit = _4;
        ctx.handle = _5;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _5(e) {
        return ctx.raise(e);
      }
    }
    /** 
     * same us `peel` without parameters but don't crashes 
     * if there is nothing to peel (returns undefined)
     */
    peelOpt() {
      const v = this.cur();
      if (!v || !v.enter) return null;
      return this.peel();
    }
    /** see `Level` */
    one() {
      var ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _2;
      ctx.handle = _3;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          if (_this._stack[0] !== vCloseTag) {
            a = Object.getPrototypeOf(Object.getPrototypeOf(_this)).one.call(_this);
            ctx.step = _2;
            return ctx.delegate(a);
          } else {
            return ctx.pure(null);
          }
        } catch (ex) {
          ctx.step = _3;
          return ctx.step(ex);
        }
      }

      function _2(r) {
        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _3(e) {
        return ctx.raise(e);
      }
    }
    /** see `Level` */
    sub() {
      var ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _2;
      ctx.handle = _3;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          if (_this._stack[0] !== vCloseTag) {
            a = Object.getPrototypeOf(Object.getPrototypeOf(_this)).sub.call(_this);
            ctx.step = _2;
            return ctx.delegate(a);
          } else {
            return ctx.pure(null);
          }
        } catch (ex) {
          ctx.step = _3;
          return ctx.step(ex);
        }
      }

      function _2(r) {
        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _3(e) {
        return ctx.raise(e);
      }
    }
    /** see `Level` */
    findPos(pos) {
      var ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _2;
      ctx.handle = _3;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          if (_this._stack[0] !== vCloseTag) {
            a = Object.getPrototypeOf(Object.getPrototypeOf(_this)).findPos.call(_this, pos);
            ctx.step = _2;
            return ctx.delegate(a);
          } else {
            return ctx.pure(null);
          }
        } catch (ex) {
          ctx.step = _3;
          return ctx.step(ex);
        }
      }

      function _2(r) {
        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _3(e) {
        return ctx.raise(e);
      }
    }
    /** peels `i` outputs it, outputs all inner  */
    copy(i) {
      var ctx = M.generator(),
          _this = this;

      ctx.$exit = _5;
      ctx.handle = _6;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          a = _this.peel(i);
          ctx.step = _2;
          ctx.value = a;
          return ctx;
        } catch (ex) {
          ctx.step = _6;
          return ctx.step(ex);
        }
      }

      function _2() {
        var a;

        try {
          a = _this.sub();
          ctx.step = _3;
          return ctx.delegate(a);
        } catch (ex) {
          ctx.step = _6;
          return ctx.step(ex);
        }
      }

      function _3() {
        var a;

        try {
          a = _this.leave();
          ctx.step = _4;
          return ctx.delegate(a);
        } catch (ex) {
          ctx.step = _6;
          return ctx.step(ex);
        }
      }

      function _4() {
        try {
          return ctx.pure();
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _5(r) {
        ctx.$exit = _5;
        ctx.handle = _6;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _6(e) {
        return ctx.raise(e);
      }
    }
    /** copies everything till closing of open tag `i` */
    tillClose(i) {
      var j,
          ctx = M.generator(),
          loop,
          _this = this,
          r;

      ctx.$exit = _4;
      ctx.handle = _5;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        exit: {
          try {
            loop = M.iterator(_this);
            ctx.step = _2;
            break exit;
          } catch (ex) {
            ctx.step = _5;
            return ctx.step(ex);
          }
        }

        return ctx.step();
      }

      function _2() {
        try {
          if (!(loop = loop.step()).done) {
            j = loop.value;
            ctx.step = _3;
            ctx.value = j;
            return ctx;
          } else {
            return ctx.pure(r);
          }
        } catch (ex) {
          ctx.step = _5;
          return ctx.step(ex);
        }
      }

      function _3() {
        exit: {
          try {
            if (j.value == i.value) {
              return ctx.pure(j);
            } else {
              ctx.step = _2;
              break exit;
            }
          } catch (ex) {
            ctx.step = _5;
            return ctx.step(ex);
          }
        }

        return ctx.step();
      }

      function _4(r) {
        ctx.$exit = _4;
        ctx.handle = _5;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _5(e) {
        return ctx.raise(e);
      }
    }
    /** 
     * consumes next tag if `i` is not closing one
     * crashes for debugging purposes if the tag is not closing tag of `i`
     */
    close(i) {
      if (!i.leave) {
        const j = this.take();
        assert.ok(i.value === j.value);
        return j;
      }
      return null;
    }
  };
}

const memo = new Map();

function parse(jsCode) {
  return (0, _babylon.parse)(jsCode, { sourceType: "module", plugins: ["dynamicImport"] });
}

/**
 * parses string `s` and outuputs its token stream at `pos`
 * following prefixes are supported in the string
 *  - '=' - the string is JS expression
 *  - '*' - the string is a list of statements
 *  - '>' - variable declarator
 *  - otherwise it is considered to be a statement
 * 
 * if in the code in `s` identifier starts with `$` it has special meaning
 * if next character is:
 *  - digit - it will replaces symbol of identifier by `syms` 
 *            at the corresponding position (1 based)
 *  - `$` -   keeps only one `$`
 *  - `I` -   unshifts a symbol from `syms` and replaces the symbol of 
 *            the identifier
 */
function toks(pos, s, ...syms) {
  var r,
      mod,
      js,
      b,
      _s,
      i,
      toks = M.generator(),
      loop;

  function replace(s) {
    var i,
        node,
        name,
        rest,
        replace = M.generator(),
        loop;
    replace.$exit = _4;
    replace.handle = _5;
    replace.step = _1;
    return replace.unwrap;

    function _1() {
      exit: {
        try {
          if (!syms.length) {
            replace.step = _3;
            return replace.delegate(s);
          } else {
            loop = M.iterator(s);
            replace.step = _2;
            break exit;
          }
        } catch (ex) {
          replace.step = _5;
          return replace.step(ex);
        }
      }

      return replace.step();
    }

    function _2() {
      try {
        if (!(loop = loop.step()).done) {
          i = loop.value;

          if (i.enter) {
            ({ node } = i.value);

            node.loc = node.start = node.end = null;
            if (i.type === _core.Tag.Identifier && i.value.node.name[0] === "$") {
              ({ name } = node);
              rest = name.substr(1);

              switch (rest[0]) {
                case "$":
                  {
                    node.name = rest;
                    break;
                  }
                case "I":
                  {
                    assert.ok(syms.length > 0);
                    i.value.node.name = (i.value.sym = syms.shift()).name;
                    break;
                  }
                default:
                  {
                    if (!isNaN(rest)) {
                      i.value.node.name = (i.value.sym = syms[rest - 1]).name;
                    }
                  }
              }
            }
          }
          replace.value = i;
          return replace;
        } else {
          return replace.pure();
        }
      } catch (ex) {
        replace.step = _5;
        return replace.step(ex);
      }
    }

    function _3() {
      try {
        return replace.pure();
      } catch (ex) {
        return replace.pure(ex);
      }
    }

    function _4(r) {
      replace.$exit = _4;
      replace.handle = _5;

      try {
        return replace.pure(r);
      } catch (ex) {
        return replace.pure(ex);
      }
    }

    function _5(e) {
      return replace.raise(e);
    }
  }
  toks.$exit = _5;
  toks.handle = _6;
  toks.step = _1;
  return toks.unwrap;

  function _1() {
    var a;

    exit: {
      try {
        if (Array.isArray(s)) {
          a = clone(s);
          toks.step = _2;
          return toks.delegate(a);
        } else {
          toks.step = _2;
          break exit;
        }
      } catch (ex) {
        toks.step = _6;
        return toks.step(ex);
      }
    }

    return toks.step();
  }

  function _2() {
    var a;

    exit: {
      try {
        if (s.substr != null) {
          r = memo.get(s);

          if (r == null) {
            mod = null;
            js = _s;

            switch (_s[0]) {
              case "=":
                {}
              // expression
              case "*":
                {}
              // list of statements
              case ">":
                {
                  // var declarator
                  mod = _s[0];
                  js = _s.slice(1);
                }
            }
            b = parse(js);

            assert.equal(b.type, "File");
            assert.equal(b.program.type, "Program");
            if (!mod === "*") {
              assert.equal(b.program.body.length, 1);
            }switch (mod) {
              case "=":
                {
                  assert.equal(b.program.body[0].type, "ExpressionStatement");
                  r = b.program.body[0].expression;
                  break;
                }
              case ">":
                {
                  assert.equal(b.program.body[0].type, "ExpressionStatement");
                  _s = b.program.body[0].expression;

                  assert.equal(_s.type, "AssignmentExpression");
                  r = T.variableDeclarator(_s.left, _s.right);
                  break;
                }
              case "*":
                {
                  break;
                }
              default:
                {
                  r = b.program.body[0];
                }
            }
            if (mod === "=" || mod === ">") {} else {
              if (mod === "*") {
                r = b.program.body;
              } else {
                r = b.program.body[0];
              }
            }memo.set(_s, r);
          }
          s = r;
        }if (Array.isArray(s)) {
          loop = M.iterator(s);
          toks.step = _3;
          break exit;
        } else {
          a = replace(clone((0, _core.produce)(s, pos)));
          toks.step = _4;
          return toks.delegate(a);
        }
      } catch (ex) {
        toks.step = _6;
        return toks.step(ex);
      }
    }

    return toks.step();
  }

  function _3() {
    var a;

    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;
        a = replace(clone((0, _core.produce)(i, pos)));
        return toks.delegate(a);
      } else {
        return toks.pure();
      }
    } catch (ex) {
      toks.step = _6;
      return toks.step(ex);
    }
  }

  function _4() {
    try {
      return toks.pure();
    } catch (ex) {
      return toks.pure(ex);
    }
  }

  function _5(r) {
    toks.$exit = _5;
    toks.handle = _6;

    try {
      return toks.pure(r);
    } catch (ex) {
      return toks.pure(ex);
    }
  }

  function _6(e) {
    return toks.raise(e);
  }
}

/** adds templated output functionality to iterator */
function Template(Super) {
  const templateTok = { $: ctrlTokGen,
    run(t) {
      var ctx = M.generator(),
          r;
      ctx.$exit = _3;
      ctx.handle = _4;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          a = t._tstack.shift();
          ctx.step = _2;
          return ctx.delegate(a);
        } catch (ex) {
          ctx.step = _4;
          return ctx.step(ex);
        }
      }

      function _2() {
        try {
          return ctx.pure(true);
        } catch (ex) {
          ctx.step = _4;
          return ctx.step(ex);
        }
      }

      function _3(r) {
        ctx.$exit = _3;
        ctx.handle = _4;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _4(e) {
        return ctx.raise(e);
      }
    } };
  return class Template extends Super {
    constructor(cont) {
      super(cont);
      this._tstack = [];
    }
    /** 
     * same as `toks` but extends formats with `$_`/'$E` identifier to mark 
     * a position where output stream must be stopped, the position is a 
     * result of the function, to continue emitting use `refocus`
     * to exit the template use `leave`
     * in expression statements $E focuses only the expression, while `$_` is for 
     * the statement, in other contexts they do matches the same
     */
    template(pos, node, ...syms) {
      var ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _2;
      ctx.handle = _3;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        var a;

        try {
          if (node.substr != null) {
            node = toArray(toks(pos, node, ...syms));
          }_this._stack.unshift(templateTok);
          _this._tstack.unshift(node);
          a = _this.refocus();
          ctx.step = _2;
          return ctx.delegate(a);
        } catch (ex) {
          ctx.step = _3;
          return ctx.step(ex);
        }
      }

      function _2(r) {
        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _3(e) {
        return ctx.raise(e);
      }
    }
    /** emits current template stream till next `$_`/`$E` */
    refocus() {
      var arr,
          f,
          n,
          _n,
          ctx = M.generator(),
          _this = this,
          r;

      ctx.$exit = _4;
      ctx.handle = _5;
      ctx.step = _1;
      return ctx.unwrap;

      function _1() {
        exit: {
          try {
            arr = _this._tstack[0];
            ctx.step = _2;
            break exit;
          } catch (ex) {
            ctx.step = _5;
            return ctx.step(ex);
          }
        }

        return ctx.step();
      }

      function _2() {
        var fn;

        exit: {
          try {
            if (arr.length) {
              f = arr.shift();

              if (f.enter) {
                switch (f.type) {
                  case _core.Tag.ExpressionStatement:
                    n = arr[0];

                    if (n != null && n.type === _core.Tag.Identifier && n.value.node.name === "$_") {
                      while (arr.length && arr.shift().value !== f.value) {}
                      return ctx.pure(f.pos);
                    } else {
                      ctx.step = _3;
                      fn = ctx.step;
                      break exit;
                    }

                  case _core.Tag.Identifier:
                    if (f.type === _core.Tag.Identifier) {
                      _n = f.value.node.name;

                      if (_n === "$_" || _n === "$E") {
                        while (arr.length && arr.shift().value !== f.value) {}
                        return ctx.pure(f.pos);
                      } else {
                        ctx.step = _3;
                        fn = ctx.step;
                        break exit;
                      }
                    } else {
                      ctx.step = _3;
                      fn = ctx.step;
                      break exit;
                    }

                  default:
                    ctx.step = _3;
                    fn = ctx.step;
                    break exit;
                }
              } else {
                ctx.step = _3;
                fn = ctx.step;
                break exit;
              }
            } else {
              throw new Error("next placeholder is not found");
            }
          } catch (ex) {
            ctx.step = _5;
            return ctx.step(ex);
          }
        }

        return fn();
      }

      function _3() {
        try {
          ctx.step = _2;
          ctx.value = f;
          return ctx;
        } catch (ex) {
          ctx.step = _5;
          return ctx.step(ex);
        }
      }

      function _4(r) {
        ctx.$exit = _4;
        ctx.handle = _5;

        try {
          return ctx.pure(r);
        } catch (ex) {
          return ctx.pure(ex);
        }
      }

      function _5(e) {
        return ctx.raise(e);
      }
    }
  };
}

/** 
 * returns an iterator depending on facilities requested in `opts` 
 *    opts: {input: boolean, 
 *           output: boolean,
 *           level: boolean, 
 *           peel: boolean,
 *           template: boolean}
 */
function Stream(opts) {
  if (opts == null) opts = {};
  let Iterator = opts.input || opts.level || opts.peel ? opts.arr ? ArrayLookahead : Lookahead : NoInput;
  if (opts.peel || opts.level) Iterator = Level(Iterator);
  if (opts.template || opts.output || opts.peel) Iterator = Output(Iterator);
  if (opts.peel) Iterator = WithPeel(Iterator);
  if (opts.template) Iterator = Template(Iterator);
  return Iterator;
}

const LookaheadArrStream = exports.LookaheadArrStream = Stream({ arr: true, input: true });
const LookaheadStream = exports.LookaheadStream = Stream({ input: true });
function lookahead(s) {
  // return Array.isArray(s) ? new LookaheadArrStream(s) : new LookaheadStream(s)
  return new LookaheadStream(s);
}

const LevelStream = exports.LevelStream = Stream({ level: true });
const LevelArrStream = exports.LevelArrStream = Stream({ level: true, arr: true });
function levels(s) {
  //  return Array.isArray(s) ? new LevelStream(s) : new LevelArrStream(s)
  return new LevelStream(s);
}

const AutoStream = exports.AutoStream = Stream({ peel: true, template: true });
const AutoArrStream = exports.AutoArrStream = Stream({ peel: true, template: true, arr: true });
function auto(s) {
  //  return Array.isArray(s) ? new AutoArrStream(s) : new AutoStream(s)
  return new AutoStream(s);
}

class NoInput {}
exports.NoInput = NoInput;
const OutputStream = exports.OutputStream = Template(Output(NoInput));
function output() {
  return new OutputStream();
}

/** runs iterable `s` and ignores its output, returns its result */
function skip(s) {
  const iter = s[Symbol.iterator]();
  let i = iter.next();
  if (i.done) return i.value;
  let node = i.value.node;
  let pos = node && (node.loc || node._loc);
  const babel = _opts.babel;
  const name = i.value.value.stageName;
  if (!babel) {
    for (; !(i = iter.next()).done;) {}
  } else {
    try {
      for (; !(i = iter.next()).done;) {
        const next = i.value.node;
        if (next && (next._loc || next.loc)) node = next;
      }
    } catch (e) {
      if (!e.esNode) e.esNode = node;
      throw e;
    }
  }
  return i.value;
}

/** modifies token replacing its `type` field */
function setType(i, type) {
  return { enter: i.enter, leave: i.leave, type, pos: i.pos, value: i.value };
}

/** modifies token replacing its `pos` field */
function setPos(i, pos) {
  return { enter: i.enter, leave: i.leave, type: i.type, pos, value: i.value };
}

/** changes position on a first level of Iterable `s` */
function repos(s, pos) {
  var iter,
      i,
      level,
      v,
      repos = M.generator(),
      r;
  repos.$exit = _4;
  repos.handle = _5;
  repos.step = _1;
  return repos.unwrap;

  function _1() {
    exit: {
      try {
        iter = s[Symbol.iterator]();
        i = void 0;
        level = 0;
        repos.step = _2;
        break exit;
      } catch (ex) {
        repos.step = _5;
        return repos.step(ex);
      }
    }

    return repos.step();
  }

  function _2() {
    var a;

    try {
      if (!(i = iter.next()).done) {
        v = i.value;

        if (v.enter) {
          level++;
        }a = level === 1 ? setPos(v, pos) : v;
        repos.step = _3;
        repos.value = a;
        return repos;
      } else {
        return repos.pure(i.value);
      }
    } catch (ex) {
      repos.step = _5;
      return repos.step(ex);
    }
  }

  function _3() {
    exit: {
      try {
        if (v.leave) {
          level--;
        }repos.step = _2;
        break exit;
      } catch (ex) {
        repos.step = _5;
        return repos.step(ex);
      }
    }

    return repos.step();
  }

  function _4(r) {
    repos.$exit = _4;
    repos.handle = _5;

    try {
      return repos.pure(r);
    } catch (ex) {
      return repos.pure(ex);
    }
  }

  function _5(e) {
    return repos.raise(e);
  }
}

/** probably faster version of `reposOne` if there is only one child */
function reposOne(s, pos) {
  var iter,
      i,
      p,
      reposOne = M.generator(),
      r;
  reposOne.$exit = _5;
  reposOne.handle = _6;
  reposOne.step = _1;
  return reposOne.unwrap;

  function _1() {
    exit: {
      try {
        iter = s[Symbol.iterator]();
        i = void 0;
        p = iter.next();

        if (p.done) {
          return reposOne.pure(p.value);
        } else {
          p = setPos(p.value, pos);
          reposOne.step = _2;
          break exit;
        }
      } catch (ex) {
        reposOne.step = _6;
        return reposOne.step(ex);
      }
    }

    return reposOne.step();
  }

  function _2() {
    var a;

    try {
      if (!(i = iter.next()).done) {
        reposOne.step = _3;
        reposOne.value = p;
        return reposOne;
      } else {
        a = setPos(p, pos);
        reposOne.step = _4;
        reposOne.value = a;
        return reposOne;
      }
    } catch (ex) {
      reposOne.step = _6;
      return reposOne.step(ex);
    }
  }

  function _3() {
    exit: {
      try {
        p = i.value;
        reposOne.step = _2;
        break exit;
      } catch (ex) {
        reposOne.step = _6;
        return reposOne.step(ex);
      }
    }

    return reposOne.step();
  }

  function _4() {
    try {
      return reposOne.pure(i.value);
    } catch (ex) {
      reposOne.step = _6;
      return reposOne.step(ex);
    }
  }

  function _5(r) {
    reposOne.$exit = _5;
    reposOne.handle = _6;

    try {
      return reposOne.pure(r);
    } catch (ex) {
      return reposOne.pure(ex);
    }
  }

  function _6(e) {
    return reposOne.raise(e);
  }
}

/** same as `reposOne` but works only for arrays avoiding traversal */
function reposOneArr(arr, pos) {
  arr[0] = setPos(arr[0], pos);
  arr[arr.length - 1] = setPos(arr[arr.length - 1], pos);
  return arr;
}

/** token to mark inner tag position must be changed to this tag postion */
const Subst = exports.Subst = (0, _core.symbol)("Subst", "ctrl");

/** applied `Subst` tokens */
function completeSubst(s) {
  var sl,
      completeSubst = M.generator();

  function subst(pos) {
    var i,
        _subst = M.generator(),
        loop;

    _subst.$exit = _5;
    _subst.handle = _6;
    _subst.step = _1;
    return _subst.unwrap;

    function _1() {
      exit: {
        try {
          loop = M.iterator(sl.sub());
          _subst.step = _2;
          break exit;
        } catch (ex) {
          _subst.step = _6;
          return _subst.step(ex);
        }
      }

      return _subst.step();
    }

    function _2() {
      var a, b;

      exit: {
        try {
          if (!(loop = loop.step()).done) {
            i = loop.value;

            if (i.type === Subst) {
              if (i.enter) {
                a = subst(pos);
                return _subst.delegate(a);
              } else {
                break exit;
              }
            } else {
              b = sl.peel(setPos(i, pos));
              _subst.step = _3;
              _subst.value = b;
              return _subst;
            }
          } else {
            return _subst.pure();
          }
        } catch (ex) {
          _subst.step = _6;
          return _subst.step(ex);
        }
      }

      return _subst.step();
    }

    function _3() {
      var a;

      try {
        a = walk();
        _subst.step = _4;
        return _subst.delegate(a);
      } catch (ex) {
        _subst.step = _6;
        return _subst.step(ex);
      }
    }

    function _4() {
      var a;

      try {
        a = sl.leave();
        _subst.step = _2;
        return _subst.delegate(a);
      } catch (ex) {
        _subst.step = _6;
        return _subst.step(ex);
      }
    }

    function _5(r) {
      _subst.$exit = _5;
      _subst.handle = _6;

      try {
        return _subst.pure(r);
      } catch (ex) {
        return _subst.pure(ex);
      }
    }

    function _6(e) {
      return _subst.raise(e);
    }
  }
  function walk() {
    var i,
        walk = M.generator(),
        loop;
    walk.$exit = _3;
    walk.handle = _4;
    walk.step = _1;
    return walk.unwrap;

    function _1() {
      exit: {
        try {
          loop = M.iterator(sl.sub());
          walk.step = _2;
          break exit;
        } catch (ex) {
          walk.step = _4;
          return walk.step(ex);
        }
      }

      return walk.step();
    }

    function _2() {
      var a;

      exit: {
        try {
          if (!(loop = loop.step()).done) {
            i = loop.value;

            if (i.type === Subst) {
              if (i.enter) {
                assert.ok(!i.leave);
                a = subst(i.pos);
                return walk.delegate(a);
              } else {
                break exit;
              }
            } else {
              walk.value = i;
              return walk;
            }
          } else {
            return walk.pure();
          }
        } catch (ex) {
          walk.step = _4;
          return walk.step(ex);
        }
      }

      return walk.step();
    }

    function _3(r) {
      walk.$exit = _3;
      walk.handle = _4;

      try {
        return walk.pure(r);
      } catch (ex) {
        return walk.pure(ex);
      }
    }

    function _4(e) {
      return walk.raise(e);
    }
  }
  completeSubst.$exit = _3;
  completeSubst.handle = _4;
  completeSubst.step = _1;
  return completeSubst.unwrap;

  function _1() {
    var a;

    try {
      sl = auto(s);
      a = walk();
      completeSubst.step = _2;
      return completeSubst.delegate(a);
    } catch (ex) {
      completeSubst.step = _4;
      return completeSubst.step(ex);
    }
  }

  function _2() {
    try {
      return completeSubst.pure();
    } catch (ex) {
      return completeSubst.pure(ex);
    }
  }

  function _3(r) {
    completeSubst.$exit = _3;
    completeSubst.handle = _4;

    try {
      return completeSubst.pure(r);
    } catch (ex) {
      return completeSubst.pure(ex);
    }
  }

  function _4(e) {
    return completeSubst.raise(e);
  }
}

/** 
 * same as `Array.from` but returns `s` if it is already `Array`
 */
function toArray(s) {
  if (Array.isArray(s)) return s;
  const res = [];
  result(s, res);
  return res;
}

/** 
 * same as `Array.from` but outputs the array into `buf` 
 * and returns iterable result
 */
function result(s, buf) {
  const iter = s[Symbol.iterator]();
  let i = iter.next();
  if (i.done) return i.value;
  buf.push(i.value);
  const name = i.value.value && i.value.value.stageName;
  const babel = _opts.babel;
  // for debugging purposes,
  // let the debugger catch the exception in browser
  if (!babel) {
    for (; !(i = iter.next()).done;) buf.push(i.value);
    return i.value;
  }
  try {
    for (; !(i = iter.next()).done;) buf.push(i.value);
  } catch (e) {
    if (e.esNode) throw e;
    let msg = e.origMessage = e.origMessage || e.message;
    if (name) msg += ` during ${name}`;
    let node = e.esNode || i && i.value.node; // || si._last
    if (!node || !node.loc && !node._loc) {
      msg += " (the position is approximated)";
      for (var loop = M.iterator(reverse(buf)); !(loop = loop.step()).done;) {
        const i = loop.value;
        node = i.value.node;
        if (node && (node.loc || node._loc)) {
          e.esNode = node;
          break;
        }
      }
    } else e.esNode = node;
    throw e;
  }
  return i.value;
}

/** alias of the module's `tillLevel` */
ExtIterator.prototype.tillLevel = function (level) {
  return tillLevel(level, this);
};

/**
 * values until leaving specified level
 */
function tillLevel(level, s) {
  var i,
      tillLevel = M.generator(),
      loop;
  tillLevel.$exit = _4;
  tillLevel.handle = _5;
  tillLevel.step = _1;
  return tillLevel.unwrap;

  function _1() {
    exit: {
      try {
        loop = M.iterator(s);
        tillLevel.step = _2;
        break exit;
      } catch (ex) {
        tillLevel.step = _5;
        return tillLevel.step(ex);
      }
    }

    return tillLevel.step();
  }

  function _2() {
    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;
        tillLevel.step = _3;
        tillLevel.value = i;
        return tillLevel;
      } else {
        return tillLevel.pure();
      }
    } catch (ex) {
      tillLevel.step = _5;
      return tillLevel.step(ex);
    }
  }

  function _3() {
    exit: {
      try {
        if (i.leave && s.level === level) {
          return tillLevel.pure();
        } else {
          tillLevel.step = _2;
          break exit;
        }
      } catch (ex) {
        tillLevel.step = _5;
        return tillLevel.step(ex);
      }
    }

    return tillLevel.step();
  }

  function _4(r) {
    tillLevel.$exit = _4;
    tillLevel.handle = _5;

    try {
      return tillLevel.pure(r);
    } catch (ex) {
      return tillLevel.pure(ex);
    }
  }

  function _5(e) {
    return tillLevel.raise(e);
  }
}

/** to be removed */
function toBlockBody(s) {
  var lab,
      i,
      toBlockBody = M.generator(),
      r;
  toBlockBody.$exit = _3;
  toBlockBody.handle = _4;
  toBlockBody.step = _1;
  return toBlockBody.unwrap;

  function _1() {
    var a;

    try {
      lab = s.label();
      i = s.cur();

      if (i.type === _core.Tag.BlockStatement) {
        s.peel();
        skip(s.peelTo(_core.Tag.body));
        return toBlockBody.pure(function () {
          var ctx = M.generator();
          ctx.$exit = _2;
          ctx.handle = _3;
          ctx.step = _1;
          return ctx.unwrap;

          function _1() {
            try {
              skip(lab());
              return ctx.pure();
            } catch (ex) {
              ctx.step = _3;
              return ctx.step(ex);
            }
          }

          function _2(r) {
            ctx.$exit = _2;
            ctx.handle = _3;

            try {
              return ctx.pure(r);
            } catch (ex) {
              return ctx.pure(ex);
            }
          }

          function _3(e) {
            return ctx.raise(e);
          }
        });
      } else {
        a = s.enter(_core.Tag.push, Subst);
        toBlockBody.step = _2;
        toBlockBody.value = a;
        return toBlockBody;
      }
    } catch (ex) {
      toBlockBody.step = _4;
      return toBlockBody.step(ex);
    }
  }

  function _2() {
    try {
      return toBlockBody.pure(lab);
    } catch (ex) {
      toBlockBody.step = _4;
      return toBlockBody.step(ex);
    }
  }

  function _3(r) {
    toBlockBody.$exit = _3;
    toBlockBody.handle = _4;

    try {
      return toBlockBody.pure(r);
    } catch (ex) {
      return toBlockBody.pure(ex);
    }
  }

  function _4(e) {
    return toBlockBody.raise(e);
  }
}

/** to be removed */
function inBlockBody(s, inner) {
  var lab,
      i,
      inBlockBody = M.generator();
  inBlockBody.$exit = _6;
  inBlockBody.handle = _7;
  inBlockBody.step = _1;
  return inBlockBody.unwrap;

  function _1() {
    var a;

    try {
      lab = s.label();
      i = s.cur();

      if (i.type !== _core.Tag.BlockStatement) {
        a = s.enter(_core.Tag.push, Subst);
        inBlockBody.step = _2;
        inBlockBody.value = a;
        return inBlockBody;
      } else {
        s.peel();
        skip(s.peelTo(_core.Tag.body));
        inBlockBody.step = _4;
        return inBlockBody.delegate(inner);
      }
    } catch (ex) {
      inBlockBody.step = _7;
      return inBlockBody.step(ex);
    }
  }

  function _2() {
    try {
      inBlockBody.step = _3;
      return inBlockBody.delegate(inner);
    } catch (ex) {
      inBlockBody.step = _7;
      return inBlockBody.step(ex);
    }
  }

  function _3() {
    var a;

    try {
      a = lab();
      inBlockBody.step = _5;
      return inBlockBody.delegate(a);
    } catch (ex) {
      inBlockBody.step = _7;
      return inBlockBody.step(ex);
    }
  }

  function _4() {
    try {
      skip(lab());
      return inBlockBody.pure();
    } catch (ex) {
      inBlockBody.step = _7;
      return inBlockBody.step(ex);
    }
  }

  function _5() {
    try {
      return inBlockBody.pure();
    } catch (ex) {
      return inBlockBody.pure(ex);
    }
  }

  function _6(r) {
    inBlockBody.$exit = _6;
    inBlockBody.handle = _7;

    try {
      return inBlockBody.pure(r);
    } catch (ex) {
      return inBlockBody.pure(ex);
    }
  }

  function _7(e) {
    return inBlockBody.raise(e);
  }
}

/** copies everything until first not-import statement */
function fileBody(s) {
  var fileBody = M.generator();
  fileBody.$exit = _3;
  fileBody.handle = _4;
  fileBody.step = _1;
  return fileBody.unwrap;

  function _1() {
    var a;

    try {
      a = s.till(i => i.pos === _core.Tag.body && i.type === _core.Tag.Array);
      fileBody.step = _2;
      return fileBody.delegate(a);
    } catch (ex) {
      fileBody.step = _4;
      return fileBody.step(ex);
    }
  }

  function _2() {
    var a;

    try {
      if (s.cur().type === _core.Tag.ImportDeclaration) {
        a = s.one();
        return fileBody.delegate(a);
      } else {
        return fileBody.pure();
      }
    } catch (ex) {
      fileBody.step = _4;
      return fileBody.step(ex);
    }
  }

  function _3(r) {
    fileBody.$exit = _3;
    fileBody.handle = _4;

    try {
      return fileBody.pure(r);
    } catch (ex) {
      return fileBody.pure(ex);
    }
  }

  function _4(e) {
    return fileBody.raise(e);
  }
}

function hasAnnot(node, name) {
  return node.leadingComments && node.leadingComments.length && node.leadingComments.find(v => v.value.trim() === name) !== undefined;
}

/** clones all tags and their `value` and `value.node` fields */
function clone(s) {
  var stack,
      i,
      value,
      isArray,
      clone = M.generator(),
      loop;
  clone.$exit = _3;
  clone.handle = _4;
  clone.step = _1;
  return clone.unwrap;

  function _1() {
    exit: {
      try {
        stack = [];
        loop = M.iterator(s);
        clone.step = _2;
        break exit;
      } catch (ex) {
        clone.step = _4;
        return clone.step(ex);
      }
    }

    return clone.step();
  }

  function _2() {
    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;
        value = null;

        if (i.enter) {
          stack.push(value = Object.assign({}, i.value));
          isArray = value.isArray = i.type === _core.Tag.Array;

          if (isArray) {
            value.node = value.node.concat();
          } else {
            if (value.node != null && _core.Tag[i.type] != null) {
              value.node = Object.assign({}, value.node);
              if (value.node.leadingComments != null) {
                value.node.leadingComments = value.node.leadingComments.concat();
              }if (value.node.trealingComments != null) {
                value.node.trealingComments = value.node.trealingComments.concat();
              }
            }
          }
        }
        if (i.leave) {
          value = stack.pop();
        }clone.value = { enter: i.enter, leave: i.leave, type: i.type, pos: i.pos, value };
        return clone;
      } else {
        return clone.pure();
      }
    } catch (ex) {
      clone.step = _4;
      return clone.step(ex);
    }
  }

  function _3(r) {
    clone.$exit = _3;
    clone.handle = _4;

    try {
      return clone.pure(r);
    } catch (ex) {
      return clone.pure(ex);
    }
  }

  function _4(e) {
    return clone.raise(e);
  }
}

/**
 * leaves all items un-amended until (and including) an item where 
 * `pred` is true
 */
function till(pred, s) {
  var i,
      till = M.generator(),
      loop,
      r;
  till.$exit = _4;
  till.handle = _5;
  till.step = _1;
  return till.unwrap;

  function _1() {
    exit: {
      try {
        loop = M.iterator(s);
        till.step = _2;
        break exit;
      } catch (ex) {
        till.step = _5;
        return till.step(ex);
      }
    }

    return till.step();
  }

  function _2() {
    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;
        till.step = _3;
        till.value = i;
        return till;
      } else {
        return till.pure(null);
      }
    } catch (ex) {
      till.step = _5;
      return till.step(ex);
    }
  }

  function _3() {
    exit: {
      try {
        if (pred(i)) {
          return till.pure(i);
        } else {
          till.step = _2;
          break exit;
        }
      } catch (ex) {
        till.step = _5;
        return till.step(ex);
      }
    }

    return till.step();
  }

  function _4(r) {
    till.$exit = _4;
    till.handle = _5;

    try {
      return till.pure(r);
    } catch (ex) {
      return till.pure(ex);
    }
  }

  function _5(e) {
    return till.raise(e);
  }
}
ExtIterator.prototype.till = function (pred) {
  return till(pred, this);
};

/** 
 * copies the stream until `pred` returns true for next token 
 * the found token is not consumed
 * returns true if found
 */
const find = exports.find = curry(function find(pred, s) {
  var i,
      _find = M.generator(),
      loop,
      r;

  _find.$exit = _3;
  _find.handle = _4;
  _find.step = _1;
  return _find.unwrap;

  function _1() {
    exit: {
      try {
        if (pred(s.cur())) {
          return _find.pure(true);
        } else {
          loop = M.iterator(s);
          _find.step = _2;
          break exit;
        }
      } catch (ex) {
        _find.step = _4;
        return _find.step(ex);
      }
    }

    return _find.step();
  }

  function _2() {
    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;

        if (pred(s.cur())) {
          return _find.pure(true);
        } else {
          _find.value = i;
          return _find;
        }
      } else {
        return _find.pure(false);
      }
    } catch (ex) {
      _find.step = _4;
      return _find.step(ex);
    }
  }

  function _3(r) {
    _find.$exit = _3;
    _find.handle = _4;

    try {
      return _find.pure(r);
    } catch (ex) {
      return _find.pure(ex);
    }
  }

  function _4(e) {
    return _find.raise(e);
  }
});

/** alias of module's find */
ExtIterator.prototype.find = function (pred) {
  return find(pred, this);
};

const Opts = exports.Opts = (0, _core.symbol)("Options");
const UpdateOpts = exports.UpdateOpts = (0, _core.symbol)("MergeOptions");

/** concatenates iterables in arguments into a single iterable */
function concat(...args) {
  var i,
      concat = M.generator(),
      loop;
  concat.$exit = _3;
  concat.handle = _4;
  concat.step = _1;
  return concat.unwrap;

  function _1() {
    exit: {
      try {
        loop = M.iterator(args);
        concat.step = _2;
        break exit;
      } catch (ex) {
        concat.step = _4;
        return concat.step(ex);
      }
    }

    return concat.step();
  }

  function _2() {
    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;
        return concat.delegate(i);
      } else {
        return concat.pure();
      }
    } catch (ex) {
      concat.step = _4;
      return concat.step(ex);
    }
  }

  function _3(r) {
    concat.$exit = _3;
    concat.handle = _4;

    try {
      return concat.pure(r);
    } catch (ex) {
      return concat.pure(ex);
    }
  }

  function _4(e) {
    return concat.raise(e);
  }
}

/** shares single iterator between several uses */
function share(s) {
  const i = s[Symbol.iterator]();
  return {
    [Symbol.iterator]() {
      return {
        next(v) {
          return i.next(v);
        }
      };
    }
  };
}

const wrap = exports.wrap = curry(function wrap(name, f, s) {
  var si,
      iter,
      i,
      j,
      e,
      babel,
      msg,
      node,
      _i,
      _wrap = M.generator(),
      loop,
      ex,
      r;

  _wrap.$exit = _5;
  _wrap.handle = _6;
  _wrap.step = _1;
  return _wrap.unwrap;

  function _1() {
    exit: {
      try {
        si = auto(s);
        iter = f(si)[Symbol.iterator]();
        i = void 0;
        _wrap.step = _2;
        break exit;
      } catch (ex) {
        _wrap.step = _6;
        return _wrap.step(ex);
      }
    }

    return _wrap.step();
  }

  function _2() {
    exit: {
      _wrap.handle = _7;

      try {
        j = void 0;
        _wrap.step = _3;
        break exit;
      } catch (ex) {
        _wrap.step = _7;
        return _wrap.step(ex);
      }
    }

    return _wrap.step();
  }

  function _3() {
    try {
      if (!(j = iter.next()).done) {
        i = j.value;
        _wrap.value = i;
        return _wrap;
      } else {
        return _wrap.pure(j.value);
      }
    } catch (ex) {
      _wrap.step = _7;
      return _wrap.step(ex);
    }
  }

  function _4() {
    _wrap.$exit = _5;
    _wrap.handle = _6;

    try {
      e = ex;
      babel = _opts.babel;

      if (babel != null) {
        msg = `${e.message} during ${name}`;
        node = e.esNode || i && i.value.node;
        // || si._last
        if (!node || !node.loc && !node._loc) {
          for (loop = M.iterator(si); !(loop = loop.step()).done;) {
            _i = loop.value;

            node = _i.value.node;
            if (node && (node.loc || node._loc)) {
              throw babel.root.hub.file.buildCodeFrameError(node, msg);
            }
          }
          node = babel.root.node;
          e.esNode = node;
        }
        throw babel.root.hub.file.buildCodeFrameError(node, msg);
      }
      throw e;
    } catch (ex) {
      _wrap.step = _6;
      return _wrap.step(ex);
    }
  }

  function _5(r) {
    _wrap.$exit = _5;
    _wrap.handle = _6;

    try {
      return _wrap.pure(r);
    } catch (ex) {
      return _wrap.pure(ex);
    }
  }

  function _6(e) {
    return _wrap.raise(e);
  }

  function _7(a) {
    _wrap.step = _4;
    ex = a;
    return _wrap.step();
  }
});

/** marks all next exceptions with `name`, eagerly consumes the former stage */
function stage(name, s) {
  const res = toArray(s);
  res[0].value.stageName = name;
  return res;
}

/** to be removed */
const checkpointLazy = exports.checkpointLazy = curry(function checkpointLazy(name, s) {
  var babel,
      iter,
      last,
      i,
      j,
      e,
      node,
      _checkpointLazy = M.generator(),
      ex,
      r;

  _checkpointLazy.$exit = _4;
  _checkpointLazy.handle = _5;
  _checkpointLazy.step = _1;
  return _checkpointLazy.unwrap;

  function _1() {
    exit: {
      try {
        babel = _opts.babel;
        iter = s[Symbol.iterator]();
        last = void 0;
        i = void 0;
        _checkpointLazy.step = _2;
        break exit;
      } catch (ex) {
        _checkpointLazy.step = _5;
        return _checkpointLazy.step(ex);
      }
    }

    return _checkpointLazy.step();
  }

  function _2() {
    _checkpointLazy.handle = _6;

    try {
      j = iter.next();

      i = j.value;
      if (j.done) {
        return _checkpointLazy.pure(i);
      } else {
        if (i.enter && i.value.node != null && i.value.node.loc != null) {
          last = i.value.node;
        }_checkpointLazy.value = i;
        return _checkpointLazy;
      }
    } catch (ex) {
      _checkpointLazy.step = _6;
      return _checkpointLazy.step(ex);
    }
  }

  function _3() {
    _checkpointLazy.$exit = _4;
    _checkpointLazy.handle = _5;

    try {
      e = ex;

      if (babel != null) {
        node = e.esNode || i && i.value.node || last || babel.root.node;

        throw babel.root.hub.file.buildCodeFrameError(node, `${e.message} during ${name}`);
      }
      throw e;
    } catch (ex) {
      _checkpointLazy.step = _5;
      return _checkpointLazy.step(ex);
    }
  }

  function _4(r) {
    _checkpointLazy.$exit = _4;
    _checkpointLazy.handle = _5;

    try {
      return _checkpointLazy.pure(r);
    } catch (ex) {
      return _checkpointLazy.pure(ex);
    }
  }

  function _5(e) {
    return _checkpointLazy.raise(e);
  }

  function _6(a) {
    _checkpointLazy.step = _3;
    ex = a;
    return _checkpointLazy.step();
  }
});

/** to be removed */
const checkpoint = exports.checkpoint = curry(function (name, s) {
  return [...checkpointLazy(name, s)];
});

/**
 * babel plugin visitor methods, typically to be applied only to Program node
 */
const babelBridge = exports.babelBridge = curry(function babelBridge(pass, path, state) {
  const optSave = _opts;
  _opts = Object.assign({ args: Object.assign({}, state.opts),
    file: Object.assign(state.file.opts),
    babel: { root: path, state } }, _opts);
  try {
    pass((0, _core.produce)({ type: "File", program: path.node }));
  } catch (e) {
    throw path.hub.file.buildCodeFrameError(e.esNode, e.message);
  }
  _opts = optSave;
});

function babelPreset(pass) {
  return {
    plugins: [function (t) {
      return {
        visitor: {
          Program(path, state) {
            path.skip();
            babelBridge(i => (0, _core.consume)(pass(i)), path, state);
          }
        }
      };
    }]
  };
}

const transform = exports.transform = curryN(2, function transform(pass, ast, opts) {
  const optSave = _opts;
  Object.assign(_opts = {}, { args: {}, file: {}, babel: false }, opts);
  try {
    return (0, _core.consume)(pass((0, _core.produce)(ast))).top;
  } finally {
    _opts = optSave;
  }
});

/**
 * copies input stream to output and returns it as array
 */
function tee(s, buf) {
  var i,
      tee = M.generator(),
      loop,
      r;
  tee.$exit = _4;
  tee.handle = _5;
  tee.step = _1;
  return tee.unwrap;

  function _1() {
    exit: {
      try {
        if (buf == null) {
          buf = [];
        }loop = M.iterator(s);
        tee.step = _2;
        break exit;
      } catch (ex) {
        tee.step = _5;
        return tee.step(ex);
      }
    }

    return tee.step();
  }

  function _2() {
    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;
        tee.step = _3;
        tee.value = i;
        return tee;
      } else {
        return tee.pure(buf);
      }
    } catch (ex) {
      tee.step = _5;
      return tee.step(ex);
    }
  }

  function _3() {
    exit: {
      try {
        buf.push(i);
        tee.step = _2;
        break exit;
      } catch (ex) {
        tee.step = _5;
        return tee.step(ex);
      }
    }

    return tee.step();
  }

  function _4(r) {
    tee.$exit = _4;
    tee.handle = _5;

    try {
      return tee.pure(r);
    } catch (ex) {
      return tee.pure(ex);
    }
  }

  function _5(e) {
    return tee.raise(e);
  }
}

/** generates error referring with `node` position */
ExtIterator.prototype.error = function (msg, node) {
  if (this._name != null) msg += " during " + this._name;
  const e = new SyntaxError(msg);
  if (node) e.esOrigNode = node;
  if (!node || !node._loc && !node.loc) {
    for (var loop = M.iterator(this); !(loop = loop.step()).done;) {
      const i = loop.value;
      node = i.value.node;
      if (node && (node.loc || node._loc)) {
        e.esNode = node;
        return e;
      }
    }
  }
  return e;
};

const makeExpr = exports.makeExpr = (0, _core.symbol)("makeExpr");
const makeStmt = exports.makeStmt = (0, _core.symbol)("makeStmt");

function makeExprPass(s) {
  s = auto(s);
  function subst(pos) {
    var t,
        subst = M.generator();
    subst.$exit = _5;
    subst.handle = _6;
    subst.step = _1;
    return subst.unwrap;

    function _1() {
      var a;

      try {
        t = s.peel();
        a = s.enter(pos, t.type, t.value);
        subst.step = _2;
        subst.value = a;
        return subst;
      } catch (ex) {
        subst.step = _6;
        return subst.step(ex);
      }
    }

    function _2() {
      var a;

      try {
        a = walk();
        subst.step = _3;
        return subst.delegate(a);
      } catch (ex) {
        subst.step = _6;
        return subst.step(ex);
      }
    }

    function _3() {
      var a;

      try {
        a = s.leave();
        subst.step = _4;
        return subst.delegate(a);
      } catch (ex) {
        subst.step = _6;
        return subst.step(ex);
      }
    }

    function _4() {
      try {
        skip(s.leave());
        return subst.pure();
      } catch (ex) {
        subst.step = _6;
        return subst.step(ex);
      }
    }

    function _5(r) {
      subst.$exit = _5;
      subst.handle = _6;

      try {
        return subst.pure(r);
      } catch (ex) {
        return subst.pure(ex);
      }
    }

    function _6(e) {
      return subst.raise(e);
    }
  }
  function toExpr(pos) {
    var j,
        ti,
        lab,
        _toExpr = M.generator();

    _toExpr.$exit = _19;
    _toExpr.handle = _20;
    _toExpr.step = _1;
    return _toExpr.unwrap;

    function _1() {
      var a, b, c, d;

      try {
        j = s.curLev();

        if (j == null) {
          return _toExpr.pure();
        } else {
          if (j.type === makeExpr || j.type === makeStmt) {
            a = s.peel(j);
            _toExpr.step = _2;
            _toExpr.value = a;
            return _toExpr;
          } else {
            ti = (0, _core.typeInfo)(j);

            if (ti.block) {
              b = s.enter(pos, _core.Tag.CallExpression);
              _toExpr.step = _4;
              _toExpr.value = b;
              return _toExpr;
            } else {
              if (ti.stmt) {
                c = s.enter(pos, _core.Tag.CallExpression);
                _toExpr.step = _10;
                _toExpr.value = c;
                return _toExpr;
              } else {
                if (ti.expr) {
                  d = subst(pos);
                  _toExpr.step = _18;
                  return _toExpr.delegate(d);
                } else {
                  throw new Error("internal: cannot convert to expression");
                }
              }
            }
          }
        }
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _2() {
      var a;

      try {
        a = toExpr(pos);
        _toExpr.step = _3;
        return _toExpr.delegate(a);
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _3() {
      var a;

      try {
        a = s.leave();
        _toExpr.step = _18;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _4() {
      var a;

      try {
        a = s.enter(_core.Tag.callee, _core.Tag.ArrowFunctionExpression);
        _toExpr.step = _5;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _5() {
      var a;

      try {
        a = s.tok(_core.Tag.params, _core.Tag.Array);
        _toExpr.step = _6;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _6() {
      var a;

      try {
        a = subst(_core.Tag.body);
        _toExpr.step = _7;
        return _toExpr.delegate(a);
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _7() {
      var a;

      try {
        a = s.leave();
        _toExpr.step = _8;
        return _toExpr.delegate(a);
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _8() {
      var a;

      try {
        a = s.tok(_core.Tag.arguments, _core.Tag.Array);
        _toExpr.step = _9;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _9() {
      var a;

      try {
        a = s.leave();
        _toExpr.step = _18;
        return _toExpr.delegate(a);
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _10() {
      var a;

      try {
        lab = s.label();
        a = s.enter(_core.Tag.callee, _core.Tag.ArrowFunctionExpression);
        _toExpr.step = _11;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _11() {
      var a;

      try {
        a = s.tok(_core.Tag.params, _core.Tag.Array);
        _toExpr.step = _12;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _12() {
      var a;

      try {
        a = s.enter(_core.Tag.body, _core.Tag.BlockStatement);
        _toExpr.step = _13;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _13() {
      var a;

      try {
        a = s.enter(_core.Tag.body, _core.Tag.Array);
        _toExpr.step = _14;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _14() {
      var a;

      try {
        a = subst(_core.Tag.push);
        _toExpr.step = _15;
        return _toExpr.delegate(a);
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _15() {
      var a;

      try {
        a = lab();
        _toExpr.step = _16;
        return _toExpr.delegate(a);
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _16() {
      var a;

      try {
        a = s.tok(_core.Tag.arguments, _core.Tag.Array);
        _toExpr.step = _17;
        _toExpr.value = a;
        return _toExpr;
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _17() {
      var a;

      try {
        a = s.leave();
        _toExpr.step = _18;
        return _toExpr.delegate(a);
      } catch (ex) {
        _toExpr.step = _20;
        return _toExpr.step(ex);
      }
    }

    function _18() {
      try {
        return _toExpr.pure();
      } catch (ex) {
        return _toExpr.pure(ex);
      }
    }

    function _19(r) {
      _toExpr.$exit = _19;
      _toExpr.handle = _20;

      try {
        return _toExpr.pure(r);
      } catch (ex) {
        return _toExpr.pure(ex);
      }
    }

    function _20(e) {
      return _toExpr.raise(e);
    }
  }
  function toStmt(pos) {
    var j,
        ti,
        _toStmt = M.generator();

    _toStmt.$exit = _7;
    _toStmt.handle = _8;
    _toStmt.step = _1;
    return _toStmt.unwrap;

    function _1() {
      var a, b, c;

      try {
        j = s.curLev();

        if (j == null) {
          return _toStmt.pure();
        } else {
          if (j.type === makeExpr || j.type === makeStmt) {
            a = s.peel(j);
            _toStmt.step = _2;
            _toStmt.value = a;
            return _toStmt;
          } else {
            ti = (0, _core.symInfo)(j.type);

            if (ti.stmt || ti.block) {
              b = subst(pos);
              _toStmt.step = _6;
              return _toStmt.delegate(b);
            } else {
              c = s.enter(pos, _core.Tag.ExpressionStatement);
              _toStmt.step = _4;
              _toStmt.value = c;
              return _toStmt;
            }
          }
        }
      } catch (ex) {
        _toStmt.step = _8;
        return _toStmt.step(ex);
      }
    }

    function _2() {
      var a;

      try {
        a = toStmt(pos);
        _toStmt.step = _3;
        return _toStmt.delegate(a);
      } catch (ex) {
        _toStmt.step = _8;
        return _toStmt.step(ex);
      }
    }

    function _3() {
      var a;

      try {
        a = s.leave();
        _toStmt.step = _6;
        _toStmt.value = a;
        return _toStmt;
      } catch (ex) {
        _toStmt.step = _8;
        return _toStmt.step(ex);
      }
    }

    function _4() {
      var a;

      try {
        a = subst(_core.Tag.expression);
        _toStmt.step = _5;
        return _toStmt.delegate(a);
      } catch (ex) {
        _toStmt.step = _8;
        return _toStmt.step(ex);
      }
    }

    function _5() {
      var a;

      try {
        a = s.leave();
        _toStmt.step = _6;
        return _toStmt.delegate(a);
      } catch (ex) {
        _toStmt.step = _8;
        return _toStmt.step(ex);
      }
    }

    function _6() {
      try {
        return _toStmt.pure();
      } catch (ex) {
        return _toStmt.pure(ex);
      }
    }

    function _7(r) {
      _toStmt.$exit = _7;
      _toStmt.handle = _8;

      try {
        return _toStmt.pure(r);
      } catch (ex) {
        return _toStmt.pure(ex);
      }
    }

    function _8(e) {
      return _toStmt.raise(e);
    }
  }
  function walk(pos) {
    var i,
        walk = M.generator(),
        loop;
    walk.$exit = _3;
    walk.handle = _4;
    walk.step = _1;
    return walk.unwrap;

    function _1() {
      exit: {
        try {
          loop = M.iterator(s.sub());
          walk.step = _2;
          break exit;
        } catch (ex) {
          walk.step = _4;
          return walk.step(ex);
        }
      }

      return walk.step();
    }

    function _2() {
      var a, b, fn;

      exit: {
        try {
          if (!(loop = loop.step()).done) {
            i = loop.value;

            switch (i.type) {
              case makeExpr:
                if (i.enter) {
                  pos = pos || i.pos;
                  a = toExpr(pos);
                  return walk.delegate(a);
                } else {
                  fn = walk.step;
                  break exit;
                }

              case makeStmt:
                if (i.enter) {
                  pos = pos || i.pos;
                  b = toStmt(pos);
                  return walk.delegate(b);
                } else {
                  fn = walk.step;
                  break exit;
                }

              default:
                walk.value = i;
                return walk;

            }
          } else {
            return walk.pure();
          }
        } catch (ex) {
          walk.step = _4;
          return walk.step(ex);
        }
      }

      return fn();
    }

    function _3(r) {
      walk.$exit = _3;
      walk.handle = _4;

      try {
        return walk.pure(r);
      } catch (ex) {
        return walk.pure(ex);
      }
    }

    function _4(e) {
      return walk.raise(e);
    }
  }
  return walk();
}

/**
 * detects if there it is a block statement with only a single result, and removes it
 */
function removeEmptyBlocks(si) {
  const s = auto(si);
  function walk(sw) {
    var i,
        lab,
        buf,
        fi,
        one,
        _walk = M.generator(),
        loop;

    _walk.$exit = _8;
    _walk.handle = _9;
    _walk.step = _1;
    return _walk.unwrap;

    function _1() {
      exit: {
        try {
          loop = M.iterator(sw);
          _walk.step = _2;
          break exit;
        } catch (ex) {
          _walk.step = _9;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _2() {
      exit: {
        try {
          if (!(loop = loop.step()).done) {
            i = loop.value;

            if (i.enter && i.type === _core.Tag.BlockStatement) {
              lab = s.label();
              buf = [s.peel(i), ...s.peelTo(_core.Tag.body)];

              if (s.curLev() == null) {
                _walk.step = _3;
                return _walk.delegate(buf);
              } else {
                fi = i.value.fieldInfo;

                if (fi.expr || fi.stmt) {
                  one = [...walk(s.one())];

                  if (s.curLev() != null || fi.expr && fi.block && (0, _core.typeInfo)(one[0]).stmt) {
                    _walk.step = _4;
                    return _walk.delegate(buf);
                  } else {
                    one[0].pos = one[one.length - 1].pos = i.pos;
                    one[0].value.fieldInfo = fi;
                    _walk.step = _7;
                    return _walk.delegate(one);
                  }
                } else {
                  break exit;
                }
              }
            } else {
              _walk.value = i;
              return _walk;
            }
          } else {
            return _walk.pure();
          }
        } catch (ex) {
          _walk.step = _9;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _3() {
      var a;

      try {
        a = lab();
        _walk.step = _2;
        return _walk.delegate(a);
      } catch (ex) {
        _walk.step = _9;
        return _walk.step(ex);
      }
    }

    function _4() {
      try {
        _walk.step = _5;
        return _walk.delegate(one);
      } catch (ex) {
        _walk.step = _9;
        return _walk.step(ex);
      }
    }

    function _5() {
      var a;

      try {
        a = walk(s.sub());
        _walk.step = _6;
        return _walk.delegate(a);
      } catch (ex) {
        _walk.step = _9;
        return _walk.step(ex);
      }
    }

    function _6() {
      var a;

      try {
        a = lab();
        _walk.step = _2;
        return _walk.delegate(a);
      } catch (ex) {
        _walk.step = _9;
        return _walk.step(ex);
      }
    }

    function _7() {
      exit: {
        try {
          skip(lab());
          _walk.step = _2;
          break exit;
        } catch (ex) {
          _walk.step = _9;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _8(r) {
      _walk.$exit = _8;
      _walk.handle = _9;

      try {
        return _walk.pure(r);
      } catch (ex) {
        return _walk.pure(ex);
      }
    }

    function _9(e) {
      return _walk.raise(e);
    }
  }
  return walk(s);
}

function adjustFieldTypeImpl(s) {
  var adjustFieldTypeImpl = M.generator(),
      r;

  function subst(pos, i) {
    var subst = M.generator();
    subst.$exit = _5;
    subst.handle = _6;
    subst.step = _1;
    return subst.unwrap;

    function _1() {
      var a, b;

      try {
        if (i.leave) {
          a = s.tok(pos, i.type, i.value);
          subst.step = _4;
          subst.value = a;
          return subst;
        } else {
          b = s.peel(setPos(i, pos));
          subst.step = _2;
          subst.value = b;
          return subst;
        }
      } catch (ex) {
        subst.step = _6;
        return subst.step(ex);
      }
    }

    function _2() {
      var a;

      try {
        a = walk();
        subst.step = _3;
        return subst.delegate(a);
      } catch (ex) {
        subst.step = _6;
        return subst.step(ex);
      }
    }

    function _3() {
      var a;

      try {
        a = s.leave();
        subst.step = _4;
        return subst.delegate(a);
      } catch (ex) {
        subst.step = _6;
        return subst.step(ex);
      }
    }

    function _4() {
      try {
        return subst.pure();
      } catch (ex) {
        return subst.pure(ex);
      }
    }

    function _5(r) {
      subst.$exit = _5;
      subst.handle = _6;

      try {
        return subst.pure(r);
      } catch (ex) {
        return subst.pure(ex);
      }
    }

    function _6(e) {
      return subst.raise(e);
    }
  }
  function walk() {
    var i,
        fi,
        ti,
        lab,
        walk = M.generator(),
        loop;
    walk.$exit = _27;
    walk.handle = _28;
    walk.step = _1;
    return walk.unwrap;

    function _1() {
      exit: {
        try {
          loop = M.iterator(s.sub());
          walk.step = _2;
          break exit;
        } catch (ex) {
          walk.step = _28;
          return walk.step(ex);
        }
      }

      return walk.step();
    }

    function _2() {
      var a, b, c, d, e, f, fn;

      exit: {
        try {
          if (!(loop = loop.step()).done) {
            i = loop.value;

            if (i.enter) {
              fi = i.value.fieldInfo || {};
              ti = (0, _core.typeInfo)(i);

              if (i.type === _core.Tag.Array && fi.array || i.type !== _core.Tag.Array && !ti.stmt && !ti.block && !ti.expr || fi.stmt && ti.stmt || fi.expr && ti.expr || fi.block && ti.block || ti.decl && fi.decl) {
                walk.value = i;
                return walk;
              } else {
                if (i.type === _core.Tag.Array) {
                  if (fi.block || fi.stmt) {
                    a = s.enter(i.pos, _core.Tag.BlockStatement);
                    walk.step = _3;
                    walk.value = a;
                    return walk;
                  } else {
                    if (fi.expr) {
                      b = s.enter(i.pos, _core.Tag.SequenceExpression);
                      walk.step = _5;
                      walk.value = b;
                      return walk;
                    } else {
                      walk.value = i;
                      return walk;
                    }
                  }
                } else {
                  if (fi.block) {
                    lab = s.label();
                    c = s.enter(i.pos, _core.Tag.BlockStatement);
                    walk.step = _7;
                    walk.value = c;
                    return walk;
                  } else {
                    if (fi.stmt && ti.expr) {
                      if (i.value.result) {
                        d = s.enter(i.pos, _core.Tag.ReturnStatement);
                        walk.step = _12;
                        walk.value = d;
                        return walk;
                      } else {
                        e = s.enter(i.pos, _core.Tag.ExpressionStatement);
                        walk.step = _14;
                        walk.value = e;
                        return walk;
                      }
                    } else {
                      if (fi.expr) {
                        f = s.enter(i.pos, _core.Tag.CallExpression);
                        walk.step = _16;
                        walk.value = f;
                        return walk;
                      } else {
                        walk.step = _26;
                        fn = walk.step;
                        break exit;
                      }
                    }
                  }
                }
              }
            } else {
              walk.step = _26;
              fn = walk.step;
              break exit;
            }
          } else {
            return walk.pure();
          }
        } catch (ex) {
          walk.step = _28;
          return walk.step(ex);
        }
      }

      return fn();
    }

    function _3() {
      var a;

      try {
        a = subst(_core.Tag.body, i);
        walk.step = _4;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _4() {
      var a;

      try {
        a = s.leave();
        walk.step = _2;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _5() {
      var a;

      try {
        a = subst(_core.Tag.expressions, i);
        walk.step = _6;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _6() {
      var a;

      try {
        a = s.leave();
        walk.step = _2;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _7() {
      var a;

      try {
        a = s.enter(_core.Tag.body, _core.Tag.Array);
        walk.step = _8;
        walk.value = a;
        return walk;
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _8() {
      var a, b, c;

      try {
        if (ti.expr) {
          if (i.value.result) {
            a = s.enter(_core.Tag.push, _core.Tag.ReturnStatement);
            walk.step = _9;
            walk.value = a;
            return walk;
          } else {
            b = s.enter(_core.Tag.push, _core.Tag.ExpressionStatement);
            walk.step = _10;
            walk.value = b;
            return walk;
          }
        } else {
          assert.ok(ti.stmt);
          c = subst(_core.Tag.push, i);
          walk.step = _11;
          return walk.delegate(c);
        }
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _9() {
      var a;

      try {
        a = subst(_core.Tag.argument, i);
        walk.step = _11;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _10() {
      var a;

      try {
        a = subst(_core.Tag.expression, i);
        walk.step = _11;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _11() {
      var a;

      try {
        a = lab();
        walk.step = _2;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _12() {
      var a;

      try {
        a = subst(_core.Tag.argument, i);
        walk.step = _13;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _13() {
      var a;

      try {
        a = s.leave();
        walk.step = _2;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _14() {
      var a;

      try {
        a = subst(_core.Tag.expression, i);
        walk.step = _15;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _15() {
      var a;

      try {
        a = s.leave();
        walk.step = _2;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _16() {
      var a;

      try {
        a = s.enter(_core.Tag.callee, _core.Tag.ArrowFunctionExpression, { node: { expression: true } });
        walk.step = _17;
        walk.value = a;
        return walk;
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _17() {
      var a;

      try {
        a = s.tok(_core.Tag.params, _core.Tag.Array);
        walk.step = _18;
        walk.value = a;
        return walk;
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _18() {
      var a, b;

      try {
        if (!ti.block) {
          a = s.enter(_core.Tag.body, _core.Tag.BlockStatement);
          walk.step = _19;
          walk.value = a;
          return walk;
        } else {
          b = subst(_core.Tag.body, i);
          walk.step = _23;
          return walk.delegate(b);
        }
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _19() {
      var a;

      try {
        a = s.enter(_core.Tag.body, _core.Tag.Array);
        walk.step = _20;
        walk.value = a;
        return walk;
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _20() {
      var a;

      try {
        a = subst(_core.Tag.push, i);
        walk.step = _21;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _21() {
      var a;

      try {
        a = s.leave();
        walk.step = _22;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _22() {
      var a;

      try {
        a = s.leave();
        walk.step = _23;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _23() {
      var a;

      try {
        a = s.leave();
        walk.step = _24;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _24() {
      var a;

      try {
        a = s.tok(_core.Tag.arguments, _core.Tag.Array);
        walk.step = _25;
        walk.value = a;
        return walk;
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _25() {
      var a;

      try {
        a = s.leave();
        walk.step = _2;
        return walk.delegate(a);
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _26() {
      try {
        walk.step = _2;
        walk.value = i;
        return walk;
      } catch (ex) {
        walk.step = _28;
        return walk.step(ex);
      }
    }

    function _27(r) {
      walk.$exit = _27;
      walk.handle = _28;

      try {
        return walk.pure(r);
      } catch (ex) {
        return walk.pure(ex);
      }
    }

    function _28(e) {
      return walk.raise(e);
    }
  }
  adjustFieldTypeImpl.$exit = _5;
  adjustFieldTypeImpl.handle = _6;
  adjustFieldTypeImpl.step = _1;
  return adjustFieldTypeImpl.unwrap;

  function _1() {
    var a;

    try {
      s = auto(s);
      a = s.peel();
      adjustFieldTypeImpl.step = _2;
      adjustFieldTypeImpl.value = a;
      return adjustFieldTypeImpl;
    } catch (ex) {
      adjustFieldTypeImpl.step = _6;
      return adjustFieldTypeImpl.step(ex);
    }
  }

  function _2() {
    var a;

    try {
      a = walk();
      adjustFieldTypeImpl.step = _3;
      return adjustFieldTypeImpl.delegate(a);
    } catch (ex) {
      adjustFieldTypeImpl.step = _6;
      return adjustFieldTypeImpl.step(ex);
    }
  }

  function _3() {
    var a;

    try {
      a = s.leave();
      adjustFieldTypeImpl.step = _4;
      return adjustFieldTypeImpl.delegate(a);
    } catch (ex) {
      adjustFieldTypeImpl.step = _6;
      return adjustFieldTypeImpl.step(ex);
    }
  }

  function _4() {
    try {
      return adjustFieldTypeImpl.pure(walk());
    } catch (ex) {
      adjustFieldTypeImpl.step = _6;
      return adjustFieldTypeImpl.step(ex);
    }
  }

  function _5(r) {
    adjustFieldTypeImpl.$exit = _5;
    adjustFieldTypeImpl.handle = _6;

    try {
      return adjustFieldTypeImpl.pure(r);
    } catch (ex) {
      return adjustFieldTypeImpl.pure(ex);
    }
  }

  function _6(e) {
    return adjustFieldTypeImpl.raise(e);
  }
}

/**
 * for expr/stmt if field type is different to actual value assigned 
 * it tries to change the value's type
 */
const adjustFieldType = exports.adjustFieldType = pipe(_core.resetFieldInfo, adjustFieldTypeImpl);

/** like `adjustFieldType` but with some simplifications */
const adjustFieldTypeSimple = exports.adjustFieldTypeSimple = pipe(_core.resetFieldInfo, removeEmptyBlocks, adjustFieldTypeImpl);

function cons(a, s) {
  var cons = M.generator();
  cons.$exit = _4;
  cons.handle = _5;
  cons.step = _1;
  return cons.unwrap;

  function _1() {
    try {
      cons.step = _2;
      cons.value = a;
      return cons;
    } catch (ex) {
      cons.step = _5;
      return cons.step(ex);
    }
  }

  function _2() {
    try {
      cons.step = _3;
      return cons.delegate(s);
    } catch (ex) {
      cons.step = _5;
      return cons.step(ex);
    }
  }

  function _3() {
    try {
      return cons.pure();
    } catch (ex) {
      return cons.pure(ex);
    }
  }

  function _4(r) {
    cons.$exit = _4;
    cons.handle = _5;

    try {
      return cons.pure(r);
    } catch (ex) {
      return cons.pure(ex);
    }
  }

  function _5(e) {
    return cons.raise(e);
  }
}

/** 
 * lookahead: returns first element and a stream with same elements
 *    
 *   `Iterable<A> -> [A, Iterable<A>]`  
 */
function la(s) {
  const i = s[Symbol.iterator]();
  const v = i.next().value;
  return [v, cons(v, i)];
}

/** 
 * runs `fn` passing `value.opts` of the stream's first output
 * if its resulting value is positive applies it to the original
 * stream or returns it unchanged
 *
 *     (Opts -> Toks?) -> Toks -> Toks
 */
const select = exports.select = curry(function select(fn, s) {
  const [h, sn] = la(s);
  const pass = fn(h.value.opts);
  return pass ? pass(sn) : sn;
});

/** 
 * if `pred` returns true on `opts` field of some first element
 * wraps transforms `s` using `pass` otherwise returns s unchanged
 *
 *     (Value -> boolean) -> (Toks -> Toks) -> (Toks -> Toks)? -> Toks -> Toks
 */
const enableIf = exports.enableIf = curryN(2, function enableIf(pred, t, e, s) {
  if (s == null) {
    if (e != null && typeof e !== "function") {
      s = e;
      e = i => i;
    }
  }
  if (e == null) e = _i => _i;
  const f = function (s) {
    const [h, sn] = la(s);
    if (!h.value.opts) h.value.opts = _opts;
    return pred(h.value) ? t(sn) : e(sn);
  };
  return s != null ? f(s) : f;
});

const packed = exports.packed = (0, _core.symbol)("packed");

/** packs not interested tokens into `packed` node */
const pack = exports.pack = curry(function pack(pred, s) {
  var buf,
      i,
      _pack = M.generator(),
      loop;

  _pack.$exit = _6;
  _pack.handle = _7;
  _pack.step = _1;
  return _pack.unwrap;

  function _1() {
    exit: {
      try {
        buf = [];
        loop = M.iterator(s);
        _pack.step = _2;
        break exit;
      } catch (ex) {
        _pack.step = _7;
        return _pack.step(ex);
      }
    }

    return _pack.step();
  }

  function _2() {
    var a, b, fn;

    exit: {
      try {
        if (!(loop = loop.step()).done) {
          i = loop.value;

          if (i.type === packed) {
            buf.push(...i.value.node);
            fn = _pack.step;
            break exit;
          } else {
            if (pred(i)) {
              if (buf.length) {
                a = (0, _core.tok)(packed, packed, { node: buf });
                _pack.step = _3;
                _pack.value = a;
                return _pack;
              } else {
                _pack.step = _4;
                fn = _pack.step;
                break exit;
              }
            } else {
              buf.push(i);
              fn = _pack.step;
              break exit;
            }
          }
        } else {
          if (buf.length) {
            b = (0, _core.tok)(packed, packed, { node: buf });
            _pack.step = _5;
            _pack.value = b;
            return _pack;
          } else {
            return _pack.pure();
          }
        }
      } catch (ex) {
        _pack.step = _7;
        return _pack.step(ex);
      }
    }

    return fn();
  }

  function _3() {
    exit: {
      try {
        buf = [];
        _pack.step = _4;
        break exit;
      } catch (ex) {
        _pack.step = _7;
        return _pack.step(ex);
      }
    }

    return _pack.step();
  }

  function _4() {
    try {
      _pack.step = _2;
      _pack.value = i;
      return _pack;
    } catch (ex) {
      _pack.step = _7;
      return _pack.step(ex);
    }
  }

  function _5() {
    try {
      return _pack.pure();
    } catch (ex) {
      return _pack.pure(ex);
    }
  }

  function _6(r) {
    _pack.$exit = _6;
    _pack.handle = _7;

    try {
      return _pack.pure(r);
    } catch (ex) {
      return _pack.pure(ex);
    }
  }

  function _7(e) {
    return _pack.raise(e);
  }
});

/** unpacks all `packed` nodes */
function unpack(s) {
  var i,
      unpack = M.generator(),
      loop;
  unpack.$exit = _3;
  unpack.handle = _4;
  unpack.step = _1;
  return unpack.unwrap;

  function _1() {
    exit: {
      try {
        loop = M.iterator(s);
        unpack.step = _2;
        break exit;
      } catch (ex) {
        unpack.step = _4;
        return unpack.step(ex);
      }
    }

    return unpack.step();
  }

  function _2() {
    var a;

    exit: {
      try {
        if (!(loop = loop.step()).done) {
          i = loop.value;

          if (i.type === packed) {
            if (i.enter) {
              a = i.value.node;
              return unpack.delegate(a);
            } else {
              break exit;
            }
          } else {
            unpack.value = i;
            return unpack;
          }
        } else {
          return unpack.pure();
        }
      } catch (ex) {
        unpack.step = _4;
        return unpack.step(ex);
      }
    }

    return unpack.step();
  }

  function _3(r) {
    unpack.$exit = _3;
    unpack.handle = _4;

    try {
      return unpack.pure(r);
    } catch (ex) {
      return unpack.pure(ex);
    }
  }

  function _4(e) {
    return unpack.raise(e);
  }
}

const filter = exports.filter = curry(function filter(pred, s) {
  var i,
      _filter = M.generator(),
      loop;

  _filter.$exit = _3;
  _filter.handle = _4;
  _filter.step = _1;
  return _filter.unwrap;

  function _1() {
    exit: {
      try {
        loop = M.iterator(s);
        _filter.step = _2;
        break exit;
      } catch (ex) {
        _filter.step = _4;
        return _filter.step(ex);
      }
    }

    return _filter.step();
  }

  function _2() {
    exit: {
      try {
        if (!(loop = loop.step()).done) {
          i = loop.value;

          if (pred(i)) {
            _filter.value = i;
            return _filter;
          } else {
            break exit;
          }
        } else {
          return _filter.pure();
        }
      } catch (ex) {
        _filter.step = _4;
        return _filter.step(ex);
      }
    }

    return _filter.step();
  }

  function _3(r) {
    _filter.$exit = _3;
    _filter.handle = _4;

    try {
      return _filter.pure(r);
    } catch (ex) {
      return _filter.pure(ex);
    }
  }

  function _4(e) {
    return _filter.raise(e);
  }
});

const flatMap = exports.flatMap = curry(function flatMap(act, s) {
  var i,
      _flatMap = M.generator(),
      loop;

  _flatMap.$exit = _3;
  _flatMap.handle = _4;
  _flatMap.step = _1;
  return _flatMap.unwrap;

  function _1() {
    exit: {
      try {
        loop = M.iterator(s);
        _flatMap.step = _2;
        break exit;
      } catch (ex) {
        _flatMap.step = _4;
        return _flatMap.step(ex);
      }
    }

    return _flatMap.step();
  }

  function _2() {
    var a;

    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;
        a = act(i);
        return _flatMap.delegate(a);
      } else {
        return _flatMap.pure();
      }
    } catch (ex) {
      _flatMap.step = _4;
      return _flatMap.step(ex);
    }
  }

  function _3(r) {
    _flatMap.$exit = _3;
    _flatMap.handle = _4;

    try {
      return _flatMap.pure(r);
    } catch (ex) {
      return _flatMap.pure(ex);
    }
  }

  function _4(e) {
    return _flatMap.raise(e);
  }
});

const map = exports.map = curry(function map(fun, s) {
  var i,
      _map = M.generator(),
      loop;

  _map.$exit = _3;
  _map.handle = _4;
  _map.step = _1;
  return _map.unwrap;

  function _1() {
    exit: {
      try {
        loop = M.iterator(s);
        _map.step = _2;
        break exit;
      } catch (ex) {
        _map.step = _4;
        return _map.step(ex);
      }
    }

    return _map.step();
  }

  function _2() {
    var a;

    try {
      if (!(loop = loop.step()).done) {
        i = loop.value;
        a = fun(i);
        _map.value = a;
        return _map;
      } else {
        return _map.pure();
      }
    } catch (ex) {
      _map.step = _4;
      return _map.step(ex);
    }
  }

  function _3(r) {
    _map.$exit = _3;
    _map.handle = _4;

    try {
      return _map.pure(r);
    } catch (ex) {
      return _map.pure(ex);
    }
  }

  function _4(e) {
    return _map.raise(e);
  }
});

const forEach = exports.forEach = curry(function forEach(act, s) {
  for (var loop = M.iterator(s); !(loop = loop.step()).done;) {
    const i = loop.value;
    act(i);
  }
});

/**
 * reverse array iterator
 */
function reverse(arr) {
  arr = toArray(arr);
  let i = arr.length;
  return {
    [Symbol.iterator]() {
      return {
        next() {
          return i === 0 ? { value: null, done: true } : { value: arr[--i], done: false };
        }
      };
    }
  };
}

/** 
 * adds value `v` to an Array in a map `m` value by `k`
 * creates the Array if needed 
 */
function mapPush(m, k, v) {
  let l = m.get(k);
  if (l == null) m.set(k, l = []);
  l.push(v);
}

/** 
 * adds value `v` to a Set in a map `m` value by `k`
 * creates the Set if needed 
 */
function mapAdd(m, k, v) {
  let l = m.get(k);
  if (l == null) m.set(k, l = new Set());
  l.add(v);
}

function groupWith(i /*:Iterable<[K,V]>*/
, accum /*:(w:W,v:V,k?:K) => W*/
, init /*::?: (k?:K) => W*/)
/*: Map<K,W>*/{
  const ret /*:Map<K,W>*/ = new Map();
  if (init == null) init = () => null;
  for (var loop = M.iterator(i); !(loop = loop.step()).done;) {
    const [k, v] = loop.value;
    const l = ret.get(k);
    ret.set(k, l == null ? accum(init(k), v, k) : accum(l, v, k));
  }
  return ret;
}

function group(i /*:Iterable<[K,V]>*/) /*: Map<K,V[]>*/{
  const ret /*:Map<K,V[]>*/ = new Map();
  for (var loop = M.iterator(i); !(loop = loop.step()).done;) {
    const [k, v] = loop.value;
    mapPush(ret, k, v);
  }return ret;
  // return groupWith<K,V,V[]>(i,(w:V[],v:V) => (w.push(v),w), () => [])
}

function groupUniq(i /*:Iterable<[K,V]>*/) /*: Map<K,Set<V>>*/{
  const ret /*:Map<K,V[]>*/ = new Map();
  for (var loop = M.iterator(i); !(loop = loop.step()).done;) {
    const [k, v] = loop.value;
    mapAdd(ret, k, v);
  }return ret;
  // return groupWith<K,V,V[]>(i,(v:V,w:V[]) => (w.add(v),w), () => new Set())
}

/** a postprocess pass for any pass removing some expressions */
const cleanEmptyExprs = exports.cleanEmptyExprs = pipe(function markLastSubExpr(si) {
  const s = auto(si);
  function walk() {
    var i,
        inner,
        _walk = M.generator(),
        loop;

    _walk.$exit = _4;
    _walk.handle = _5;
    _walk.step = _1;
    return _walk.unwrap;

    function _1() {
      exit: {
        try {
          loop = M.iterator(s.sub());
          _walk.step = _2;
          break exit;
        } catch (ex) {
          _walk.step = _5;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _2() {
      try {
        if (!(loop = loop.step()).done) {
          i = loop.value;
          _walk.step = _3;
          _walk.value = i;
          return _walk;
        } else {
          return _walk.pure();
        }
      } catch (ex) {
        _walk.step = _5;
        return _walk.step(ex);
      }
    }

    function _3() {
      exit: {
        try {
          if (i.type === _core.Tag.SequenceExpression) {
            inner = [...walk()];

            if (inner.length > 2) {
              inner[inner.length - 2].value.last = true;
            }_walk.step = _2;
            return _walk.delegate(inner);
          } else {
            _walk.step = _2;
            break exit;
          }
        } catch (ex) {
          _walk.step = _5;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _4(r) {
      _walk.$exit = _4;
      _walk.handle = _5;

      try {
        return _walk.pure(r);
      } catch (ex) {
        return _walk.pure(ex);
      }
    }

    function _5(e) {
      return _walk.raise(e);
    }
  }
  return walk();
}, function cleanEmptyExprs(si) {
  const s = auto(si);
  function canSkip(inner) {
    if (!inner.length) return true;
    const t = inner[0];
    if (t.type === _core.Tag.Null || t.value.canIgnore) return true;
    return false;
  }
  function walk(sw, ignore) {
    var i,
        j,
        buf,
        nxt,
        inner,
        k,
        _inner,
        _walk = M.generator(),
        loop,
        _loop;

    _walk.$exit = _12;
    _walk.handle = _13;
    _walk.step = _1;
    return _walk.unwrap;

    function _1() {
      exit: {
        try {
          loop = M.iterator(sw);
          _walk.step = _2;
          break exit;
        } catch (ex) {
          _walk.step = _13;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _2() {
      var a, b, fn;

      exit: {
        try {
          if (!(loop = loop.step()).done) {
            i = loop.value;

            if (i.enter) {
              switch (i.type) {
                case _core.Tag.SequenceExpression:
                  j = s.take();
                  buf = [];

                  for (nxt = void 0; (nxt = s.curLev()) != null;) {
                    inner = [...walk(s.one(), !nxt.value.last || ignore)];

                    if (nxt.value.last && !ignore || !canSkip(inner)) {
                      buf.push(inner);
                    }
                  }

                  switch (buf.length) {
                    case 1:
                      a = reposOneArr(buf[0], i.pos);
                      _walk.step = _3;
                      return _walk.delegate(a);

                    case 0:
                      _walk.step = _3;
                      fn = _walk.step;
                      break exit;

                    default:
                      _walk.step = _4;
                      _walk.value = i;
                      return _walk;

                  }

                case _core.Tag.ExpressionStatement:
                  _inner = [...walk(s.one(), true)];

                  if (canSkip(_inner)) {
                    if (i.pos !== _core.Tag.push) {
                      b = s.tok(i.pos, _core.Tag.Null);
                      _walk.step = _8;
                      _walk.value = b;
                      return _walk;
                    } else {
                      _walk.step = _8;
                      fn = _walk.step;
                      break exit;
                    }
                  } else {
                    _walk.step = _9;
                    _walk.value = i;
                    return _walk;
                  }

                default:
                  ignore = false;
                  _walk.step = _11;
                  fn = _walk.step;
                  break exit;
              }
            } else {
              _walk.step = _11;
              fn = _walk.step;
              break exit;
            }
          } else {
            return _walk.pure();
          }
        } catch (ex) {
          _walk.step = _13;
          return _walk.step(ex);
        }
      }

      return fn();
    }

    function _3() {
      exit: {
        try {
          s.close(j);
          s.close(i);
          _walk.step = _2;
          break exit;
        } catch (ex) {
          _walk.step = _13;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _4() {
      try {
        _walk.step = _5;
        _walk.value = j;
        return _walk;
      } catch (ex) {
        _walk.step = _13;
        return _walk.step(ex);
      }
    }

    function _5() {
      exit: {
        try {
          _loop = M.iterator(buf);
          _walk.step = _6;
          break exit;
        } catch (ex) {
          _walk.step = _13;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _6() {
      var a;

      try {
        if (!(_loop = _loop.step()).done) {
          k = _loop.value;
          return _walk.delegate(k);
        } else {
          a = s.close(j);
          _walk.step = _7;
          _walk.value = a;
          return _walk;
        }
      } catch (ex) {
        _walk.step = _13;
        return _walk.step(ex);
      }
    }

    function _7() {
      var a;

      try {
        a = s.close(i);
        _walk.step = _2;
        _walk.value = a;
        return _walk;
      } catch (ex) {
        _walk.step = _13;
        return _walk.step(ex);
      }
    }

    function _8() {
      exit: {
        try {
          s.close(i);
          _walk.step = _2;
          break exit;
        } catch (ex) {
          _walk.step = _13;
          return _walk.step(ex);
        }
      }

      return _walk.step();
    }

    function _9() {
      try {
        _walk.step = _10;
        return _walk.delegate(_inner);
      } catch (ex) {
        _walk.step = _13;
        return _walk.step(ex);
      }
    }

    function _10() {
      var a;

      try {
        a = s.close(i);
        _walk.step = _2;
        _walk.value = a;
        return _walk;
      } catch (ex) {
        _walk.step = _13;
        return _walk.step(ex);
      }
    }

    function _11() {
      try {
        _walk.step = _2;
        _walk.value = i;
        return _walk;
      } catch (ex) {
        _walk.step = _13;
        return _walk.step(ex);
      }
    }

    function _12(r) {
      _walk.$exit = _12;
      _walk.handle = _13;

      try {
        return _walk.pure(r);
      } catch (ex) {
        return _walk.pure(ex);
      }
    }

    function _13(e) {
      return _walk.raise(e);
    }
  }
  return walk(s);
}, _core.removeNulls);

/**
 * similar to Ramda.pipe, likely less efficient, 
 * but Ramda dependency can be dropped
 */
function pipe() {
  var args = arguments;
  return function pipeImpl(cur) {
    for (let i = 0, len = args.length; i < len; ++i) cur = args[i](cur);
    return cur;
  };
}

/**
 * similar to Ramda.curryN, likely less efficient, 
 * but Ramda dependency can be dropped
 */
function curryN(num, fun, trace) {
  function step(args) {
    return function curryStep() {
      const next = args.concat(Array.from(arguments));
      return next.length >= num ? fun.apply(undefined, next) : step(next);
    };
  }
  return step([]);
}

/**
 * similar to Ramda.curry, likely less efficient, 
 * but Ramda dependency can be dropped
 */
function curry(fun, trace) {
  return curryN(fun.length, fun, trace);
}