"use strict";

exports.__esModule = true;
exports.output = exports.fin = exports.toConsole = exports.color = exports.convertCtrl = undefined;
exports.markNodeType = markNodeType;
exports.toStr = toStr;
exports.setComment = setComment;
exports.setEndComment = setEndComment;
exports.copyComment = copyComment;

var _kit = require("./kit");

var Kit = _interopRequireWildcard(_kit);

var _scope = require("./scope");

var _trace = require("./trace");

var Trace = _interopRequireWildcard(_trace);

var _core = require("./core");

var _babelGenerator = require("babel-generator");

var _babelGenerator2 = _interopRequireDefault(_babelGenerator);

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

var _babelTypes = require("babel-types");

var T = _interopRequireWildcard(_babelTypes);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

const fs = require("fs");

const BROWSER_DEBUG = Trace.BROWSER_DEBUG;

function* markNodeType(s) {
  for (const i of s) {
    if (i.enter) {
      const ti = (0, _core.typeInfo)(i);
      if (ti.kind === "ctrl") (i.value.comments || (i.value.comments = [])).unshift({ txt: (0, _core.symName)(i.type), style: styles.nodetype });else if (i.value.comments != null && i.value.comments.length) {
        (i.value.comments || (i.value.comments = [])).unshift({ txt: (0, _core.symName)(i.type).match(/[A-Z]/g).join(""),
          style: styles.nodetype });
      }
    }
    yield i;
  }
}

const convertCtrl = exports.convertCtrl = Kit.pipe(_core.resetFieldInfo, markNodeType, Array.from, function convertCtrl(s) {
  s = Kit.auto(s);
  function* walk(sw) {
    for (const i of sw) {
      if (i.enter) {
        const fld = i.value.fieldInfo || {},
              ti = (0, _core.typeInfo)(i);
        if (ti.kind === "ctrl") {
          s.peel(i);
          const lab = s.label();
          const nm = ti.name;
          const stmt = !fld.expr;
          yield s.enter(i.pos, stmt ? _core.Tag.BlockStatement : _core.Tag.SequenceExpression, { comments: i.value.comments });
          yield s.enter(stmt ? _core.Tag.body : _core.Tag.expressions, _core.Tag.Array);
          if (!i.leave) {
            for (let j; (j = s.curLev()) != null;) {
              yield s.enter(_core.Tag.push, Kit.Subst);
              yield* walk(s.one());
              yield* s.leave();
            }
          }
          yield* lab();
          Kit.skip(s.leave());
          continue;
        }
      }
      yield i;
    }
  }
  return walk(s);
}, Kit.completeSubst, Array.from, Kit.adjustFieldType);

const color = exports.color = BROWSER_DEBUG ? function* color(s) {
  for (const i of s) {
    if (i.enter && i.value.comments) {
      for (const j of i.value.comments) {
        if (j.style != null && (0, _core.typeInfo)(i).kind === "node") {
          j.txt = `%c${j.txt}%c`;
          j.args = [j.style, ""];
        }
      }
    }
    if (i.leave && i.value.tcomments) {
      for (const j of i.value.tcomments) {
        if (j.style != null && (0, _core.typeInfo)(i).kind === "node") {
          j.txt = `%c${j.txt}%c`;
          j.args = [j.style, ""];
        }
      }
    }
    yield i;
  }
} : s => s;

function* getArgs(s) {
  for (const i of s) {
    if (i.enter && i.value.comments) {
      for (const j of i.value.comments) {
        if (j.args) for (const k of j.args) yield k;
      }
    }
    if (i.leave && i.value.tcomments) {
      for (const j of i.value.tcomments) {
        if (j.args) for (const k of j.args) yield k;
      }
    }
  }
}

function* applyComments(s) {
  for (const i of s) {
    const node = i.value.node;
    if (i.enter && node != null) {
      if (i.value.comments != null && i.value.comments.length) {
        (node.leadingComments || (node.leadingComments = [])).push({ type: "CommentBlock",
          value: i.value.comments.map(v => v.txt).join("|") });
      }
      if (i.value.tcomments != null && i.value.tcomments.length) {
        (node.trailingComments || (node.trailingComments = [])).push({ type: "CommentBlock",
          value: i.value.tcomments.map(v => v.txt).join("|") });
      }
    }
    yield i;
  }
}

const toConsole = exports.toConsole = Kit.curry(function toConsole(tag, s) {
  s = Kit.auto(s);
  const root = s.first.value;
  let name = (0, _core.symName)(s.first.type).match(/[A-Z]/g).join("");
  if (root.funcId) name += ":" + root.funcId.id;
  if (BROWSER_DEBUG) console.groupCollapsed(`dump %c${tag} %c${name}`, "color:orange;font-size:large", "color:coral");else console.log(`dump ${tag}`);
  const col = Kit.pipe(convertCtrl, color, _scope.tempNames, Array.from, applyComments, Array.from)(s);
  const args = Array.from(getArgs(col));
  (0, _core.consume)(Trace.verify(col));
  console.log((0, _babelGenerator2.default)(col[0].value.node).code, ...args);
  if (BROWSER_DEBUG) console.groupEnd();
});

const fin = exports.fin = Kit.pipe(_core.removeNulls, _scope.tempNames, convertCtrl, Array.from, applyComments, Array.from);

function toStr(s) {
  return (0, _babelGenerator2.default)((0, _core.consume)(fin(Kit.repos(s, _core.Tag.top))).top).code;
}

const output = exports.output = Kit.curry(function dump(tag, s) {
  const sa = Kit.toArray(s);
  const sl = Kit.auto(Kit.clone(sa));
  const opts = sl.opts || {};
  let dest;
  if (BROWSER_DEBUG) {
    dest = "console";
  } else {
    dest = `dump-${tag}.js`;
    if (opts.dump) {
      if (opts.dump.substr) {
        if (opts.dump === "console") {
          dest = opts.dump;
        } else {
          dest = opts.dump + dest;
        }
      } else {
        const s = opts.dump[tag];
        if (s == null && s.substr) {
          dest = s;
        }
      }
    }
  }
  if (dest === "console") {
    toConsole(tag, sl);
  } else {
    fs.writeFileSync(dest, toStr(sl));
  }
  return sa;
});

exports.default = output;
function setComment(i, txt, style = "none") {
  style = style.substr && styles[style] || style;
  (i.value.comments || (i.value.comments = [])).push({ txt, style });
  return i;
}

function setEndComment(i, txt, style = "none") {
  style = style.substr && styles[style] || style;
  (i.value.tcomments || (i.value.tcomments = [])).push({ txt, style });
  return i;
}

function copyComment(f, t) {
  if (f.value.comments != null) (t.value.comments || (t.value.comments = [])).push(...f.value.comments);
  if (f.value.tcomments != null) (t.value.tcomments || (t.value.tcomments = [])).push(...f.value.tcomments);
  return t;
}

const styles = {
  large: "font-size:xx-large;color:orange",
  hl: "font-size:large;color:navy;",
  small: "font-size:large;color:navy;",
  smallH: "font-size:large;color:orange;font-weight:bolder",
  nodetype: "font-size:xx-small;color:green;font-weight:bolder",
  none: ""
};