"use strict";

exports.__esModule = true;
exports.run = exports.inject = exports.clean = exports.commit = exports.Placeholder = exports.Root = undefined;

var _kit = require("./kit");

var Kit = _interopRequireWildcard(_kit);

var _core = require("./core");

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

var _trace = require("./trace");

var Trace = _interopRequireWildcard(_trace);

var _dump = require("./dump");

var _dump2 = _interopRequireDefault(_dump);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

const Root = exports.Root = (0, _core.symbol)("match.root", "ctrl");
const Placeholder = exports.Placeholder = (0, _core.symbol)("match.placeholder", "ctrl");

const commit = exports.commit = Kit.pipe(Array.from, function* (s) {
  const buf = [];
  for (let i of s) {
    switch (i.type) {
      case Root:
        if (!i.value.v.match) continue;
        i = i.value.s ? (0, _core.enter)(i.pos, i.type, i.value.v) : (0, _core.leave)(i.pos, i.type, i.value.v);
        break;
      case Placeholder:
        if (!i.value.v.match) continue;
    }
    yield i;
  }
});

const clean = exports.clean = function* (s) {
  for (const i of s) {
    switch (i.type) {
      case Root:
      case Placeholder:
        break;
      default:
        yield i;
    }
  }
};

const inject = exports.inject = Kit.curry(function* matchInject(pattern, si) {
  const pats = (Array.isArray(pattern) ? pattern : [pattern]).map(i => Kit.toArray(Kit.toks(_core.Tag.top, i)));
  const starts = pats.map((i, x) => i[0].type);
  const plen = pats.length;
  const plens = pats.map(i => i.length);
  assert.equal(plens.filter(v => v === 0).length, 0);
  const s = Kit.lookahead(si);
  const activePos = [];
  const activeTok = [];
  const activePh = [];
  const activePat = [];
  let level = 0;
  for (let i of s) {
    if (i.enter) {
      level++;
    }
    let aplen = activePos.length;
    for (let p = 0; p < aplen; ++p) {
      let x = activePos[p];
      const pat = activePat[p];
      const plen = pat.length;
      const v = activeTok[p];
      const ph = activePh[p];
      if (x !== plen && x !== -1 && activePh[p] == null) {
        let j = pat[x++];
        //during generation we may neglect emitting empty arrays 
        if (j.pos !== i.pos && j.type === _core.Tag.Array && j.enter && !j.leave && pat[x].value === j.value) {
          x++;
          j = pat[x++];
        }
        if (x < plen && j.pos != i.pos && j.enter === i.enter) {
          activePos[p] = -1;
          v.match = false;
          continue;
        }
        if (j.enter) {
          switch (j.type) {
            case _core.Tag.ExpressionStatement:
              const k = pat[x];
              if (k.type === _core.Tag.Identifier) {
                let block = false;
                if (k.value.node.name[0] === "$") {
                  const ph = activePh[p] = { v, level, name: k.value.node.name.substr(1) };
                  yield (0, _core.enter)(i.pos, Placeholder, ph);
                  x++;
                  assert.equal(pat[x].value, k.value);
                  x++;
                  assert.equal(pat[x].value, j.value);
                  x++;
                  activePos[p] = x;
                  continue;
                }
              }
              break;
            case _core.Tag.Identifier:
            case _core.Tag.StringLiteral:
              let block = false;
              const name = j.value.node.name || j.value.node.value;
              if (name[0] === "$") {
                const ph = activePh[p] = { v, level, name: name.substr(1) };
                yield (0, _core.enter)(i.pos, Placeholder, ph);
                if (!j.leave) {
                  assert.equal(pat[x].value, j.value);
                  x++;
                }
                activePos[p] = x;
                continue;
              }
              break;
          }
        }
        if (j.type !== i.type || j.value.node.name !== i.value.node.name || j.value.node.value !== i.value.node.value) {
          activePos[p] = -1;
          v.match = false;
          continue;
        }
        if (i.enter) {
          assert.ok(j.enter);
          if (i.leave && !j.leave) {
            if (pat[x++].value !== j.value) {
              activePos[p] = -1;
              v.match = false;
              continue;
            }
          } else if (j.leave) {
            if (s.cur().value === i.value) s.read();
          }
        }
        activePos[p] = x;
      }
    }
    if (i.enter) {
      for (let s = 0; s < plen; ++s) {
        if (starts[s] == i.type) {
          const v = { match: null, index: s };
          yield (0, _core.tok)(i.pos, Root, { s: true, v });
          activePos.push(1);
          activeTok.push(v);
          activePat.push(pats[s]);
        }
      }
    }
    yield i;
    if (i.leave) {
      let aplen = activePos.length;
      for (let p = aplen - 1; p >= 0; --p) {
        let ph = activePh[p];
        if (ph != null) {
          if (ph.level === level) {
            yield (0, _core.leave)(i.pos, Placeholder, ph);
            ph = activePh[p] = null;
          }
        }
        if (ph == null) {
          let x = activePos[p];
          if (x >= 0 && x === activePat[p].length) {
            activePos[p] = -1;
            const v = activeTok[p];
            v.match = true;
            while (activePos[0] === -1) {
              activePos.shift();
              activePh.shift();
              activeTok.shift();
              activePat.shift();
            }
            yield (0, _core.tok)(i.pos, Root, { s: false, v });
          }
        }
      }
      level--;
    }
  }
});

const run = exports.run = Kit.curry(function (pat, si) {
  return commit(inject(pat, si));
}, true);