"use strict";

exports.__esModule = true;
exports.importSyms = importSyms;
exports.inline = inline;
exports.init = init;

var _kit = require("./kit");

var Kit = _interopRequireWildcard(_kit);

var _scope = require("./scope");

var Scope = _interopRequireWildcard(_scope);

var _core = require("./core");

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

/** 
 * injects ES6 `import` or commonjs `require` 
 * commonjs if option `modules` === 'commonjs'
 * expects following fields:
 *      type Value = Value & { 
 *        rt: { importSyms: {syms:Sym[], ns: Sym, module:string} } }
 *      type Sym = Sym & { importFrom: String } 
 */
function* importSyms(si) {
  const s = Kit.auto(si);
  const { rt } = s.first.value;
  if (!rt.importSyms || !rt.importSyms.length) {
    yield* s;
    return;
  }
  yield* Kit.fileBody(s);
  const commonjs = s.opts.modules === "commonjs";
  for (const { syms, ns, module } of rt.importSyms) {
    if (ns == null) {
      const lab = s.label();
      yield* s.template(_core.Tag.push, commonjs ? `var $_ = require("${module}")` : `import $_ from "${module}"`);
      yield s.enter(_core.Tag.id, _core.Tag.ObjectPattern);
      yield s.enter(_core.Tag.properties, _core.Tag.Array);
      for (const sym of syms) {
        yield s.enter(_core.Tag.push, _core.Tag.ObjectProperty, { node: { shorthand: true } });
        yield s.tok(_core.Tag.key, _core.Tag.Identifier, { node: { name: sym.orig } });
        yield s.tok(_core.Tag.value, _core.Tag.Identifier, { decl: true, sym });
        yield* s.leave();
      }
      yield* lab();
    } else {
      yield* s.toks(_core.Tag.push, commonjs ? `var $I = require("${module}")` : `import * as $I from "${module}"`, ns);
    }
  }
  yield* s;
}

/** 
 * inlines runtime values
 * expects: 
 *      type Value = Value & { rt: { inline: {syms:Sym[],content:string}[] } }
 * inlines JS content from all `inlineSources` looking up for 
 * symbols in `inlineSym`
 */
function* inline(si) {
  const s = Kit.auto(si);
  const { rt } = s.first.value;
  const inlines = rt.inlineSyms;
  if (!inlines || !inlines.length) {
    yield* s;
    return;
  }
  function* getBody(si) {
    const s = Kit.auto(si);
    Kit.skip(s.till(i => i.pos === _core.Tag.body && i.type === _core.Tag.Array));
    yield* s.sub();
    Kit.skip(s);
  }
  const buf = [];
  const transf = Kit.pipe(Kit.parse, _core.produce, Scope.prepare, getBody, Kit.toArray);
  for (const { syms, content } of inlines) {
    const symsMap = {};
    for (const i of syms) symsMap[i.orig] = i;
    const p = transf(content);
    buf.push(p);
    for (const i of p) {
      if (i.enter) {
        const { node } = i.value;
        node.loc = node.start = node.end = null;
        if (i.type === _core.Tag.Identifier) {
          const sym = symsMap[node.name];
          if (sym) i.value.sym = sym;else if (i.value.sym) i.value.sym.strict = false;
        }
      }
    }
  }
  yield* Kit.fileBody(s);
  for (const i of buf) yield* i;
  yield* s;
}

/** initializes empty `rt` structure */
function* init(si) {
  const [h, s] = Kit.la(si);
  h.value.rt = { importSyms: [],
    importNs: [],
    inlineSyms: [] };
  yield* s;
}