"use strict";

exports.__esModule = true;

var _core = require("../core");

var _kit = require("../kit");

var Kit = _interopRequireWildcard(_kit);

var _trace = require("../trace");

var Trace = _interopRequireWildcard(_trace);

var _dump = require("../dump");

var _dump2 = _interopRequireDefault(_dump);

var _scope = require("../scope");

var Scope = _interopRequireWildcard(_scope);

var _rt = require("../rt");

var RT = _interopRequireWildcard(_rt);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

const fs = require("fs");
const path = require("path");

/** calculates captured vars dependencies */
function calcClosCapt(si) {
  const sa = Kit.toArray(si);
  const s = Kit.auto(sa);
  const closureSym = s.first.value.closureSym = Scope.newSym("closure");
  closureSym.global = true;
  const rt = s.first.value.rt;
  if (!s.opts.noRT) rt.inlineSyms.push({ syms: [closureSym],
    content: fs.readFileSync(path.join(__dirname, "closConvRT.js"), "utf-8")
  });
  function walk(root, sw) {
    const decls = root.decls = [];
    const sym = root.closSym = Scope.newSym(root.node.id && root.node.id.name || "fn");
    const closDeps = new Set();
    function id(i) {
      const si = i.value.sym;
      if (si != null) {
        if (si === Scope.argumentsSym && !root.argumentsSym) {
          root.argumentsSym = Scope.newSym("args");
          return;
        }
        if (i.value.decl) decls.push(si);
        if (si.declScope) {
          if (si.declScope !== root) {
            (si.refScopes || (si.refScopes = new Set())).add(root);
            closDeps.add(si.declScope);
          }
        }
      }
    }
    for (const i of sw) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionDeclaration:
            id(s.cur());
            Kit.skip(s.one());
          case _core.Tag.FunctionExpression:
            walk(i.value, s.sub());
            break;
          case _core.Tag.Identifier:
            id(i);
            break;
        }
      }
    }
    root.closDeps = [...closDeps].sort((a, b) => a.closSym.num - b.closSym.num);
  }
  walk(s.first.value, s);
  return sa;
}

const selfSym = Scope.newSym("self");
const thisSym = Scope.newSym("this", true);
const globals = Scope.newSym("g");

/** replaces function calls, to call method */
function replaceCalls(si) {
  const s = Kit.auto(si);
  s.first.value.ctxSym = globals;
  function* walk(sw, decls) {
    for (const i of sw) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionDeclaration:
          case _core.Tag.FunctionExpression:
            yield i;
            i.value.ctxSym = thisSym;
            yield* walk(s.sub(), i.value.decls);
            continue;
          case _core.Tag.NewExpression:
          case _core.Tag.CallExpression:
            const j = s.curLev();
            const lab = s.label();
            const constr = i.type === _core.Tag.NewExpression;
            yield s.peel(Kit.setType(i, _core.Tag.CallExpression));
            yield s.enter(_core.Tag.callee, _core.Tag.MemberExpression);
            let sym;
            if (j.type === _core.Tag.MemberExpression) {
              s.take();
              yield s.enter(_core.Tag.object, j.type, j.value);
              const k = s.curLev();
              if (k.type === _core.Tag.Identifier || constr) {
                yield* s.one();
                sym = k.value.sym;
              } else {
                decls.push(sym = Scope.newSym("temp"));
                yield* s.template(k.pos, "=$I = $_", sym);
                yield* Kit.reposOne(walk(s.one(), decls), _core.Tag.right);
                yield* s.leave();
              }
              yield* walk(s.one(), decls);
              yield* s.leave();
              s.close(j);
            } else {
              sym = Scope.undefinedSym;
              yield* Kit.reposOne(walk(s.one()), _core.Tag.object);
            }
            yield s.tok(_core.Tag.property, _core.Tag.Identifier, { node: { name: constr ? "constr" : "call" } });
            yield* s.leave();
            yield s.peel();
            if (!constr) yield s.tok(_core.Tag.push, _core.Tag.Identifier, { sym });
            yield* walk(s.sub(), decls);
            yield* s.leave();
            yield* lab();
            continue;
        }
      }
      yield i;
    }
  }
  return walk(s, s.first.value.decls);
}

function* functToObj(si) {
  const s = Kit.auto(si);
  const hoisted = [];
  const rtSym = new Scope.newSym("RT");
  const closureSym = s.first.value.closureSym;
  function* walk(sw, root, blockHoisted) {
    function* func(i, pos) {
      const sym = i.value.closSym;
      yield* s.template(pos, "=new $I($_)", sym);
      for (const j of i.value.closDeps) {
        if (j === root) yield s.tok(_core.Tag.push, _core.Tag.Identifier, { sym: root.ctxSym });else yield* s.toks(_core.Tag.push, "=this.$I", j.closSym);
      }
      yield* s.leave();
      hoisted.push(Kit.toArray(obj(i.value, sym)));
      s.close(i);
    }
    for (const i of sw) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.BlockStatement:
            yield i;
            yield* s.peelTo(_core.Tag.body);
            const hoisted = [];
            const buf = Kit.toArray(walk(s.sub(), root, hoisted));
            for (const j of hoisted) yield* j;
            yield* buf;
            yield* s.leave();
            continue;
          case _core.Tag.ThisExpression:
            yield s.tok(i.pos, _core.Tag.Identifier, { sym: selfSym });
            s.close(i);
            continue;
          case _core.Tag.VariableDeclaration:
            const vlab = s.label();
            let pos = i.pos;
            if (i.pos === _core.Tag.left) {
              const j = s.take();
              const k = s.take();
              yield s.tok(i.pos, _core.Tag.Identifier, { sym: s.curLev().value.sym });
              Kit.skip(s.sub());
              s.close(k);
              s.close(j);
              s.close(i);
              continue;
            }
            if (pos === _core.Tag.push) {
              yield s.enter(_core.Tag.push, _core.Tag.ExpressionStatement);
              pos = _core.Tag.expression;
            }
            let sym;
            yield s.enter(pos, _core.Tag.SequenceExpression);
            yield s.enter(_core.Tag.expressions, _core.Tag.Array);
            for (const j of s.sub()) {
              if (j.enter) {
                for (const j of s.sub()) {
                  const id = Kit.toArray(s.one());
                  sym = id[0].value.sym;
                  if (s.curLev() != null) {
                    yield* s.template(_core.Tag.push, "=$I = $_", sym);
                    yield* Kit.reposOne(walk(s.one(), root, blockHoisted), _core.Tag.right);
                    yield* s.leave();
                  }
                  s.close(j);
                }
              }
            }
            yield* vlab();
            s.close(i);
            continue;
          case _core.Tag.FunctionDeclaration:
            blockHoisted.push(Kit.toArray(function* (i) {
              yield* s.template(i.pos, "$I = $_", s.curLev().value.sym);
              yield* func(i, _core.Tag.right);
              yield* s.leave();
            }(i)));
            continue;
          case _core.Tag.FunctionExpression:
            yield* func(i, i.pos);
            continue;
        }
      }
      yield i;
    }
  }
  function* obj(fun, sym) {
    const lab = s.label();
    yield* s.template(_core.Tag.push, "*function $1($_){$_} $2($1, $_)", sym, closureSym);
    for (const j of fun.closDeps) yield s.tok(_core.Tag.push, _core.Tag.Identifier, { sym: j.closSym });
    yield* s.refocus();
    for (const j of fun.closDeps) yield* s.toks(_core.Tag.push, "$1.$2 = $2", fun.ctxSym, j.closSym);
    yield* s.refocus();
    yield s.enter(_core.Tag.push, _core.Tag.FunctionExpression, fun);
    yield* walk(s.sub(), fun);
    yield* lab();
  }
  yield* Kit.fileBody(s);
  const buf = Kit.toArray(walk(s, s.first.value, hoisted));
  yield* s.toks(_core.Tag.push, "var $I = {}", s.first.value.ctxSym);
  for (const i of hoisted) yield* i;
  yield* buf;
  yield* s;
}

function substIds(si) {
  const s = Kit.auto(si);
  function* emitDecls({ decls, argumentsSym, ctxSym }) {
    const locs = [];
    const params = [];
    for (const i of decls) {
      if (i.param) {
        if (i.refScopes) params.push(i);
      } else if (!i.refScopes) locs.push(i);
    }
    yield* s.till(i => i.pos === _core.Tag.body && i.type === _core.Tag.Array);
    const lab = s.label();
    if (locs.length || argumentsSym) {
      yield s.enter(_core.Tag.push, _core.Tag.VariableDeclaration, { node: { kind: "var" } });
      yield s.enter(_core.Tag.declarations, _core.Tag.Array);
      if (argumentsSym) {
        yield s.enter(_core.Tag.push, _core.Tag.VariableDeclarator);
        yield s.tok(_core.Tag.id, _core.Tag.Identifier, { sym: argumentsSym });
        yield* s.toks(_core.Tag.init, "=Array.from(arguments).slice(1)");
        yield* s.leave();
      }
      for (const sym of locs) {
        yield s.enter(_core.Tag.push, _core.Tag.VariableDeclarator);
        yield s.tok(_core.Tag.id, _core.Tag.Identifier, { sym });
        yield* s.leave();
      }
      yield* lab();
    }
    if (params.length) {
      yield s.enter(_core.Tag.push, _core.Tag.ExpressionStatement);
      yield s.enter(_core.Tag.expression, _core.Tag.SequenceExpression);
      yield s.enter(_core.Tag.expressions, _core.Tag.Array);
      for (const i of params) yield* s.toks(_core.Tag.push, "=$1.$2 = $2", ctxSym, i);
      yield* lab();
    }
  }
  function* walk(root) {
    for (const i of s.sub()) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionExpression:
            yield i;
            yield* s.till(i => i.pos === _core.Tag.params);
            yield s.tok(_core.Tag.push, _core.Tag.Identifier, { sym: selfSym });
            yield* emitDecls(i.value);
            yield* walk(i.value);
            continue;
          case _core.Tag.File:
            yield i;
            yield* emitDecls(i.value);
            continue;
          case _core.Tag.Identifier:
            const { sym } = i.value;
            if (sym && i.pos !== _core.Tag.property) {
              if (sym === Scope.argumentsSym) {
                i.value.sym = root.argumentsSym;
              } else if (sym.refScopes) {
                if (root === sym.declScope) yield* s.toks(i.pos, "=$I.$I", root.ctxSym, sym);else yield* s.toks(i.pos, "=$I.$I.$I", root.ctxSym, sym.declScope.closSym, sym);
                s.close(i);
                continue;
              }
            }
            break;
        }
      }
      yield i;
    }
  }
  return walk(s.first.value);
}

exports.default = Kit.pipe(Scope.prepare, RT.init, calcClosCapt, replaceCalls, functToObj, Kit.toArray, substIds, RT.inline, Scope.resolve);