"use strict";

exports.__esModule = true;
exports.setModule = setModule;

var _core = require("../core");

var _kit = require("../kit");

var Kit = _interopRequireWildcard(_kit);

var _trace = require("../trace");

var Trace = _interopRequireWildcard(_trace);

var _dump = require("../dump");

var _dump2 = _interopRequireDefault(_dump);

var _scope = require("../scope");

var Scope = _interopRequireWildcard(_scope);

var _ramda = require("ramda");

var R = _interopRequireWildcard(_ramda);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function* inlineRT(si) {
  var s = Kit.auto(si);
  var rt = s.first.rt;

  var buf = [];
  var symsMap = {};
  function* getBody(si) {
    var s = Kit.auto(si);
    Kit.skip(s.till(i => i.pos === _core.Tag.body && i.type === _core.Tag.Array));
    yield* s.sub();
    Kit.skip(s);
  }
  for (var i of rt.syms) {
    symsMap[i.orig] = i;
  }for (var _i of rt.sources) {
    var p = R.pipe(Kit.parse, _core.produce, Scope.prepare, getBody, Kit.toArray)(_i);
    buf.push(p);
    for (var _i2 of p) {
      if (_i2.enter) {
        var node = _i2.value.node;

        node.loc = node.start = node.end = null;
        if (_i2.type === _core.Tag.Identifier) {
          var sym = symsMap[node.name];
          if (sym) _i2.value.sym = sym;else if (_i2.value.sym) _i2.value.sym.strict = false;
        }
      }
    }
  }
  yield* s.till(i => i.pos === _core.Tag.body && i.type === _core.Tag.Array);
  while (s.cur().type === _core.Tag.ImportDeclaration) {
    yield* s.one();
  }for (var _i3 of buf) {
    yield* _i3;
  }yield* s;
}

var modules = {};

function setModule(name, code) {
  modules[name] = code;
}

function* setRT(si) {
  var _Kit$la = Kit.la(si),
      h = _Kit$la[0],
      s = _Kit$la[1];

  var closureSym = Scope.newSym("closure");
  h.rt = { sources: Object.values(modules),
    syms: [closureSym],
    closureSym };
  yield* s;
}

/** calculates captured vars dependencies */
function calcClosCapt(si) {
  var sa = Kit.toArray(si);
  var s = Kit.auto(sa);
  function walk(root, sw) {
    var decls = root.decls = [];
    var sym = root.closSym = Scope.newSym(root.node.id && root.node.id.name || "fn");
    var closCapt = new Set();
    var closDeps = new Set();
    function id(i) {
      var si = i.value.sym;
      if (si != null) {
        if (si === Scope.argumentsSym && !root.argumentsSym) {
          root.argumentsSym = Scope.newSym("args");
          return;
        }
        if (i.value.decl) decls.push(si);
        if (si.declScope) {
          if (si.declScope !== root) {
            (si.refScopes || (si.refScopes = new Set())).add(root);
            closDeps.add(si.declScope);
          } else closCapt.add(si);
        }
      }
    }
    for (var i of sw) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionDeclaration:
            id(s.cur());
            Kit.skip(s.one());
          case _core.Tag.FunctionExpression:
            walk(i.value, s.sub());
            break;
          case _core.Tag.Identifier:
            id(i);
            break;
        }
      }
    }
    root.closCapt = [].concat(_toConsumableArray(closCapt)).sort((a, b) => a.num - b.num);
    root.closDeps = [].concat(_toConsumableArray(closDeps)).sort((a, b) => a.closSym.num - b.closSym.num);
  }
  walk(s.first.value, s);
  return sa;
}

/** replaces function calls, to call method */
function replaceCalls(si) {
  var s = Kit.auto(si);
  function* walk(sw, decls) {
    for (var i of sw) {
      yield i;
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionDeclaration:
          case _core.Tag.FunctionExpression:
            yield* walk(s.sub(), i.value.decls);
            break;
          case _core.Tag.NewExpression:
            var constr = true;
          case _core.Tag.CallExpression:
            var j = s.curLev();
            yield s.enter(_core.Tag.callee, _core.Tag.MemberExpression);
            var sym = void 0;
            if (j.type === _core.Tag.MemberExpression) {
              s.take();
              yield s.enter(_core.Tag.object, j.type, j.value);
              var k = s.curLev();
              if (k.type === _core.Tag.Identifier) {
                yield* s.one();
                sym = k.value.sym;
              } else {
                decls.push(sym = Scope.newSym("temp"));
                yield* s.template(k.pos, "=$I = $_", sym);
                yield* Kit.reposOne(walk(s.one(), decls), _core.Tag.right);
                yield* s.leave();
              }
              yield* walk(s.one(), decls);
              yield* s.leave();
              s.close(j);
            } else {
              sym = Scope.undefinedSym;
              yield* Kit.reposOne(walk(s.one()), _core.Tag.object);
            }
            yield s.tok(_core.Tag.property, _core.Tag.Identifier, { node: { name: constr ? "constr" : "call" } });
            yield* s.leave();
            yield s.peel();
            yield s.tok(_core.Tag.push, _core.Tag.Identifier, { sym });
            yield* walk(s.sub(), decls);
            yield* s.leave();
        }
      }
    }
  }
  return walk(s, s.first.value.decls);
}

var selfSym = Scope.newSym("self");

function* functToObj(si) {
  var s = Kit.auto(si);
  var hoisted = [];
  var blockHoisted = [];
  var rtSym = new Scope.newSym("RT");
  var closSym = s.first.rt.closureSym;
  function* walk(sw, root) {
    function* func(i, pos) {
      var sym = i.value.closSym;
      yield* s.template(pos, "=new $I($_)", sym);
      for (var j of i.value.closDeps) {
        if (j === root) yield s.tok(_core.Tag.push, _core.Tag.ThisExpression);else yield* s.toks(_core.Tag.push, "=this.$I", j.closSym);
      }
      yield* s.leave();
      hoisted.push(Kit.toArray(obj(i.value, sym)));
      s.close(i);
    }
    for (var i of sw) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.BlockStatement:
            yield i;
            yield* s.peelTo(_core.Tag.body);
            var _buf = Kit.toArray(walk(s.sub(), root));
            for (var j of blockHoisted) {
              yield* j;
            }blockHoisted.length = 0;
            yield* _buf;
            yield* s.leave();
            continue;
          case _core.Tag.ThisExpression:
            yield s.tok(i.pos, _core.Tag.Identifier, { sym: selfSym });
            s.close(i);
            continue;
          case _core.Tag.VariableDeclaration:
            var vlab = s.label();
            if (i.pos === _core.Tag.push) {
              yield s.enter(_core.Tag.push, _core.Tag.ExpressionStatement);
              yield s.enter(_core.Tag.expression, _core.Tag.SequenceExpression);
              yield s.enter(_core.Tag.expressions, _core.Tag.Array);
            }
            var sym = void 0;
            for (var _j of s.sub()) {
              if (_j.enter) {
                for (var _j2 of s.sub()) {
                  var id = Kit.toArray(s.one());
                  sym = id[0].value.sym;
                  if (s.curLev() != null) {
                    yield* s.template(i.pos, "=$I = $_", sym);
                    yield* Kit.reposOne(walk(s.one(), root), _core.Tag.right);
                    yield* s.leave();
                  }
                  s.close(_j2);
                }
              }
            }
            yield* vlab();
            if (i.pos !== _core.Tag.push) yield s.tok(i.pos, _core.Tag.Identifier, { sym });
            s.close(i);
            continue;
          case _core.Tag.FunctionDeclaration:
            blockHoisted.push(Kit.toArray(function* (i) {
              yield* s.template(i.pos, "$I = $_", s.curLev().value.sym);
              yield* func(i, _core.Tag.right);
              yield* s.leave();
            }(i)));
            continue;
          case _core.Tag.FunctionExpression:
            yield* func(i, i.pos);
            continue;
        }
      }
      yield i;
    }
  }
  function* obj(fun, sym) {
    var lab = s.label();
    yield* s.template(_core.Tag.push, "*function $1($_){$_} $2($1, $_)", sym, closSym);
    for (var j of fun.closDeps) {
      yield s.tok(_core.Tag.push, _core.Tag.Identifier, { sym: j.closSym });
    }yield* s.refocus();
    for (var _j3 of fun.closDeps) {
      yield* s.toks(_core.Tag.push, "this.$1 = $1", _j3.closSym);
    }yield* s.refocus();
    yield s.enter(_core.Tag.push, _core.Tag.FunctionExpression, fun);
    yield* walk(s.sub(), fun);
    yield* lab();
  }
  yield* s.till(i => i.type === _core.Tag.Array && i.pos === _core.Tag.body);
  var buf = Kit.toArray(walk(s, s.first.value));
  for (var i of hoisted) {
    yield* i;
  }for (var _i4 of blockHoisted) {
    yield* _i4;
  }yield* buf;
  yield* s;
}

function substIds(si) {
  var s = Kit.auto(si);
  function* emitDecls(_ref) {
    var decls = _ref.decls,
        argumentsSym = _ref.argumentsSym;

    decls = decls.filter(i => !i.param && !i.refScopes);
    yield* s.till(i => i.pos === _core.Tag.body && i.type === _core.Tag.Array);
    if (decls.length || argumentsSym) {
      var lab = s.label();
      yield s.enter(_core.Tag.push, _core.Tag.VariableDeclaration, { node: { kind: "var" } });
      yield s.enter(_core.Tag.declarations, _core.Tag.Array);
      if (argumentsSym) {
        yield s.enter(_core.Tag.push, _core.Tag.VariableDeclarator);
        yield s.tok(_core.Tag.id, _core.Tag.Identifier, { sym: argumentsSym });
        yield* s.toks(_core.Tag.init, "=Array.from(arguments).slice(1)");
        yield* s.leave();
      }
      for (var sym of decls) {
        yield s.enter(_core.Tag.push, _core.Tag.VariableDeclarator);
        yield s.tok(_core.Tag.id, _core.Tag.Identifier, { sym });
        yield* s.leave();
      }
      yield* lab();
    }
  }
  function* walk(root) {
    for (var i of s.sub()) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionExpression:
            yield i;
            yield* s.till(i => i.pos === _core.Tag.params);
            yield s.tok(_core.Tag.push, _core.Tag.Identifier, { sym: selfSym });
            yield* emitDecls(i.value);
            yield* walk(i.value);
            continue;
          case _core.Tag.File:
            yield i;
            yield* emitDecls(i.value);
            continue;
          case _core.Tag.Identifier:
            var sym = i.value.sym;

            if (sym && i.pos !== _core.Tag.property) {
              if (sym === Scope.argumentsSym) {
                i.value.sym = root.argumentsSym;
              } else if (sym.refScopes) {
                if (root === sym.declScope) yield* s.toks(i.pos, "=this.$I", sym);else yield* s.toks(i.pos, "=this.$I.$I", sym.declScope.closSym, sym);
                s.close(i);
                continue;
              }
            }
            break;
        }
      }
      yield i;
    }
  }
  return walk(s.first.value);
}

exports.default = R.pipe(Scope.prepare, setRT, Trace.all("INITA"), calcClosCapt, replaceCalls, Trace.all("calls"), (0, _dump2.default)("calls"), functToObj, Trace.all("REPLA"), (0, _dump2.default)("objs"), substIds, Trace.all("SubSt"), (0, _dump2.default)("SUbst"), inlineRT, Scope.resolve);